import pyPPG

import pandas as pd
from pyPPG.ppg_bm.bm_extraction import get_biomarkers

###########################################################################
######################### Get Ratios of PPG Signal ########################
###########################################################################
def get_sig_ratios(s: pyPPG.PPG, fp: pyPPG.Fiducials):
    """
    This function returns the biomarkers of Signal ratios.

    :param s: object of PPG signal
    :type s: pyPPG.PPG object
    :param fp: object of fiducial points
    :type fp: pyPPG.Fiducials object

    :return:
        - biomarkers: dictionary of biomarkers of Signal ratios
        - biomarkers_lst: list a biomarkers with name, definition and unit
    """

    biomarkers_lst = [
                    ["IPR",          "Instantaneous pulse rate, 60 / Tpi", "[%]"],
                    ["Tsys/Tdia",    "Ratio of the systolic time vs. the diastolic time", "[%]"],
                    ["Tpw25/Tpi",    "Ratio of the pulse width at 25% of the systolic peak amplitude vs. the pulse interval", "[%]"],
                    ["Tpw50/Tpi",    "Ratio of the pulse width at 50% of the systolic peak amplitude vs. the pulse interval", "[%]"],
                    ["Tpw75/Tpi",    "Ratio of the pulse width at 75% of the systolic peak amplitude vs. the pulse interval", "[%]"],
                    ["Tpw25/Tsp",    "Ratio of the pulse width at 25% of the systolic peak amplitude vs. the systolic peak time", "[%]"],
                    ["Tpw50/Tsp",    "Ratio of the pulse width at 50% of the systolic peak amplitude vs. the systolic peak time", "[%]"],
                    ["Tpw75/Tsp",    "Ratio of the pulse width at 75% of the systolic peak amplitude vs. the systolic peak time", "[%]"],
                    ["Tdw10/Tsw10",  "Ratio of the diastolic width vs. the systolic width at 10% width", "[%]"],
                    ["Tdw25/Tsw25",  "Ratio of the diastolic width vs. the systolic width at 25% width", "[%]"],
                    ["Tdw33/Tsw33",  "Ratio of the diastolic width vs. the systolic width at 33% width", "[%]"],
                    ["Tdw50/Tsw50",  "Ratio of the diastolic width vs. the systolic width at 50% width", "[%]"],
                    ["Tdw66/Tsw66",  "Ratio of the diastolic width vs. the systolic width at 66% width", "[%]"],
                    ["Tdw75/Tsw75",  "Ratio of the diastolic width vs. the systolic width at 75% width", "[%]"],
                    ["Tdw90/Tsw90",  "Ratio of the diastolic width vs. the systolic width at 90% width", "[%]"],
                    ["Tsp/Tpi",      "Ratio of the systolic peak time vs. the pulse interval", "[%]"],
                    ["Asp/Aoff",     "Ratio of the systolic peak amplitude vs. the pulse offset amplitude", "[%]"],
                    ["Adp/Asp",      "Reflection index, the ratio of the diastolic peak amplitude vs. the systolic peak amplitude", "[%]"],
                    ["IPA",          "Inflection point area, the ratio of the area under diastolic curve vs. the area under systolic curve", "[nu]"],
                    ["Tsp/Asp",      "Ratio of the systolic peak time vs. the systolic peak amplitude", "[nu]"],
                    ["Asp/deltaT",   "Stiffness index, the ratio of the systolic peak amplitude vs. the time delay", "[nu]"],
                    ["Asp/(Tpi-Tsp)","Ratio of the systolic peak amplitude vs. the difference between the pulse interval and systolic peak time ", "[nu]"],
    ]

    header = ['name', 'definition', 'unit']
    biomarkers_lst = pd.DataFrame(biomarkers_lst, columns=header)

    df, df_biomarkers = get_biomarkers(s, fp, biomarkers_lst.name)

    return df_biomarkers, biomarkers_lst



import pyPPG
from pyPPG.preproc import Preprocessing

import matplotlib.pyplot as plt
import scipy.io
import numpy as np
from dotmap import DotMap
from tkinter import filedialog
import mne
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
import os

###########################################################################
####################### Data Acquisition from Files #######################
###########################################################################
def load_data(data_path: str, start = 0, end = 0, filtering = True):
    """
    Load PPG data function load the raw PPG data.

    :param data_path: path of the PPG signal
    :type data_path: str
    :param start: beginning the of signal in sample
    :type start: int
    :param end: end of the signal in sample
    :type end: int
    :param filtering: a bool for filtering
    :type filtering: bool

    :return s: a struct of PPG signal:

        * s.start: beginning of the signal in sample
        * s.end: end of the signal in sample
        * s.v: a vector of PPG values
        * s.fs: the sampling frequency of the PPG in Hz
        * s.name: name of the record
        * s.v: 1-d array, a vector of PPG values
        * s.fs: the sampling frequency of the PPG in Hz
        * s.filt_sig: 1-d array, a vector of the filtered PPG values
        * s.filt_d1: 1-d array, a vector of the filtered PPG' values
        * s.filt_d2: 1-d array, a vector of the filtered PPG" values
        * s.filt_d3: 1-d array, a vector of the filtered PPG'" values
    """

    if data_path=="":
        sig_path = filedialog.askopenfilename(title='Select SIGNAL file', filetypes=[("Input Files", ".mat .csv .edf .pkl .txt")])
    else:
        sig_path=data_path

    if sig_path.rfind('/')>0:
        start_c = sig_path.rfind('/')+1
    else:
        start_c = sig_path.rfind('\\')

    stop_c=sig_path.rfind('.')
    rec_name=sig_path[start_c:stop_c]

    try:
        sig_format=sig_path[len(sig_path)-sig_path[::-1].index('.'):]
    except:
        print('Invalid signal path!')

    if sig_format=='mat':
        input_sig = scipy.io.loadmat(sig_path)
        sig = np.float64(np.squeeze(input_sig.get("Data")))[0:]
        try:
            fs = np.squeeze(input_sig.get("Fs"))
        except:
            fs = 100
            print('The default sampling frequency is 100 Hz for .mat.')
    elif sig_format=='csv':
        input_sig = np.loadtxt(sig_path, delimiter=',').astype(int)
        sig = input_sig
        fs = 75
        print('The default sampling frequency is 75 Hz for .csv.')
    elif sig_format=='txt':
        try:
            input_sig = np.loadtxt(sig_path, delimiter='\t').astype(int)
        except:
            try:
                input_sig = np.loadtxt(sig_path, delimiter=' ').astype(int)
            except:
                print('ERROR! The data separator is not supported for .txt.')
        sig = input_sig
        fs = 1000
        print('The default sampling frequency is 1 kHz for .txt.')
    elif sig_format == 'edf':
        input_sig = mne.io.read_raw_edf(sig_path)
        sig = -input_sig['Pleth'][0][0]
        try:
            fs = int(np.round(input_sig.info['sfreq']))
        except:
            fs = 256
            print('The default sampling frequency is 256 Hz for .edf.')

    s = DotMap()

    s.start = start
    if start<end:
        s.end = end
    else:
        s.end = len(sig)

    s.v=sig[s.start:s.end]
    s.fs=fs
    s.name=rec_name
    s.filt_sig, s.filt_d1, s.filt_d2, s.filt_d3 = Preprocessing(s, filtering = True)

    return s

###########################################################################
########################### Plot Fiducial points ##########################
###########################################################################
def plot_fiducials(s: pyPPG.PPG, fp: pyPPG.Fiducials, savingfolder: str):
    """
    Plot fiducial points of the filtered PPG signal.

    :param s: a struct of PPG signal
    :type s: pyPPG.PPG object
    :param fp: a struct of fiducial points
    :type fp: pyPPG.Fiducials object
    :param savingfolder: location of the saved figure
    """

    fig = plt.figure(figsize=(20, 12))
    ax1 = plt.subplot(411)
    plt.plot(s.filt_sig, 'k', label=None)
    ax2 = plt.subplot(412, sharex=ax1)
    plt.plot(s.filt_d1, 'k', label=None)
    ax3 = plt.subplot(413, sharex=ax2)
    plt.plot(s.filt_d2, 'k', label=None)
    ax4 = plt.subplot(414, sharex=ax3)
    plt.plot(s.filt_d3, 'k', label=None)
    fig.subplots_adjust(hspace=0, wspace=0)

    marker = ['o', 's', 's','o', 'o', 's', 'o', 'o', 's', 'o', 's', 'o', 's', 'o', 's']
    color = ['r', 'b', 'g','m', 'r', 'b', 'g', 'r', 'b', 'g', 'm', 'c', 'k', 'r', 'b']

    fid_names = ('sp', 'on', 'dn','dp', 'u', 'v', 'w', 'a', 'b', 'c', 'd', 'e', 'f', 'p1', 'p2')
    ylabe_names = ['PPG', 'PPG\'', 'PPG\'\'', 'PPG\'\'\'']
    s_type = ['filt_sig', 'filt_sig', 'filt_sig','filt_sig', 'filt_d1', 'filt_d1', 'filt_d1', 'filt_d2', 'filt_d2', 'filt_d2', 'filt_d2', 'filt_d2', 'filt_d2', 'filt_d3','filt_d3']

    str_sig = 0
    end_sig = len(s.filt_sig)
    len_sig=end_sig-str_sig
    step_small = 1
    step_big = step_small * 5

    major_ticks_names = range(0, int(len_sig/s.fs),step_big)
    len_ticks_names=len(major_ticks_names)
    major_diff=len_sig/len_ticks_names
    minor_diff = len_sig / len_ticks_names / step_big
    major_ticks = np.arange(str_sig, end_sig, major_diff)
    minor_ticks = np.arange(str_sig, end_sig, minor_diff)

    for n in fid_names:
        ind = fid_names.index(n)
        if s_type[ind][-1] == 'g':
            plt_num = 1
            plt.subplot(411)
            plt.title(s.name, fontsize=20)
        else:
            plt_num = int(s_type[ind][-1]) + 1


        ax = plt.subplot(4, 1, plt_num)
        tmp_pnt=eval("fp." + n + ".values")
        tmp_pnt=tmp_pnt[~np.isnan(tmp_pnt)].astype(int)
        tmp_sig=eval("s." + s_type[ind])
        exec("plt.scatter(tmp_pnt, tmp_sig[tmp_pnt], s=60,linewidth=2, marker = marker[ind],facecolors='none', color=color[ind], label=n)")
        plt.ylabel(ylabe_names[plt_num - 1], fontsize=20)

        exec("plt.xlim([str_sig,end_sig])")
        leg = plt.legend(loc='upper right', fontsize=20, ncol=2,facecolor="orange",frameon=True)
        for text in leg.get_texts():
            text.set_weight('bold')

        ax.set_xticks(major_ticks)
        ax.set_xticks(minor_ticks, minor=True)

        ax.grid(which='both')
        ax.grid(which='minor', alpha=0.2,axis='x')
        ax.grid(which='major', alpha=1,axis='x')

        plt.yticks([])

    plt.xlabel('Time [s]', fontsize=20)
    plt.xticks(major_ticks,major_ticks_names, fontsize=20)
    plt.show()

    canvas = FigureCanvas(fig)
    tmp_dir=savingfolder+os.sep+'PPG_Figures'+os.sep
    if not os.path.exists(tmp_dir):
        os.mkdir(tmp_dir)

    canvas.print_png((tmp_dir+'%s_btwn_%s-%s.png') % (s.name,s.start,s.end))
    print('Figure has been saved in the "'+savingfolder+'".')

###########################################################################
################################# Save Data ###############################
###########################################################################

def save_data(s: pyPPG.PPG, fp: pyPPG.Fiducials, bm: pyPPG.Biomarkers, savingformat: str, savingfolder: str):
    """
    Save the results of the filtered PPG analysis.

    :param s: a struct of PPG signal
    :type s: pyPPG.PPG object
    :type s: DotMap
    :param fp: a struct of fiducial points
    :type fp: pyPPG.Fiducial object
    :param bm: a dictionary of biomarkers
    :type bm: pyPPG.Biomarkers object
    :param savingformat: file format of the saved date, the provided file formats .mat and .csv
    :type savingformat: str
    :param savingfolder: location of the saved data
    :type savingfolder: str
    """

    tmp_dir = savingfolder
    os.makedirs(tmp_dir, exist_ok=True)

    temp_dirs = ['Fiducial_points', 'Biomarker_vals', 'Biomarker_stats', 'Biomarker_defs', 'PPG_struct']
    for i in temp_dirs:
        temp_dir = tmp_dir + os.sep + i + os.sep
        if not os.path.exists(temp_dir):
            os.mkdir(temp_dir)


    keys=s.__dict__.keys()
    keys_list = list(keys)
    sc=DotMap()
    for i in keys_list:
        exec('sc.'+i+' = s.'+i)

    file_name = (r'.' + os.sep + tmp_dir + os.sep + temp_dirs[4] + os.sep + s.name + '_data_btwn_%s-%s.mat')%(s.start,s.end)
    scipy.io.savemat(file_name, sc)

    BM_keys = bm.bm_vals.keys()

    if savingformat=="csv":
        file_name = (r'.'+os.sep+tmp_dir+os.sep+temp_dirs[0]+os.sep+s.name+'_'+'Fiducials_btwn_%s-%s.csv')%(s.start,s.end)
        fp.get_fp().to_csv(file_name)

        for key in BM_keys:
            file_name = (r'.'+os.sep+tmp_dir+os.sep+temp_dirs[1]+os.sep+'%s_btwn_%s-%s.csv')%(s.name+'_'+key,s.start,s.end)
            bm.bm_vals[key].to_csv(file_name,index=True,header=True)

            file_name = (r'.'+os.sep+tmp_dir+os.sep+temp_dirs[2]+os.sep+'%s_btwn_%s-%s.csv')%(s.name+'_'+key,s.start,s.end)
            bm.bm_stats[key].to_csv(file_name, index=True, header=True)

            file_name = (r'.'+os.sep+tmp_dir+os.sep+temp_dirs[3]+os.sep+'%s_btwn_%s-%s.csv')%(s.name+'_'+key,s.start,s.end)
            bm.bm_defs[key].to_csv(file_name, index=True, header=True)

    elif savingformat=="mat":
        matlab_struct = fp.get_fp().to_dict(orient='list')
        file_name = (r'.'+os.sep+tmp_dir+os.sep+temp_dirs[0]+os.sep+s.name+'_'+'Fiducials_btwn_%s-%s.mat')%(s.start,s.end)
        scipy.io.savemat(file_name,matlab_struct)

        for key in BM_keys:

            matlab_struct = bm.bm_vals[key].to_dict(orient='list')
            file_name = (r'.'+os.sep+tmp_dir+os.sep+temp_dirs[1]+os.sep+'%s_btwn_%s-%s.mat')%(s.name+'_'+key,s.start,s.end)
            scipy.io.savemat(file_name,matlab_struct)

            file_name = (r'.'+os.sep+tmp_dir+os.sep+temp_dirs[2]+os.sep+'%s_btwn_%s-%s.mat')%(s.name+'_'+key,s.start,s.end)
            scipy.io.savemat(file_name, bm.bm_stats[key])

            matlab_struct = bm.bm_defs[key].to_dict(orient='list')
            file_name = (r'.'+os.sep+tmp_dir+os.sep+temp_dirs[3]+os.sep+'%s_btwn_%s-%s.mat')%(s.name+'_'+key,s.start,s.end)
            scipy.io.savemat(file_name,matlab_struct)
    else:
        print('The file format is not suported for data saving! You can use "mat" or "csv" file formats.')

    print('Results have been saved into the "'+tmp_dir+'".')


import pyPPG

from pyPPG.ppg_bm.ppg_sig import get_ppg_sig
from pyPPG.ppg_bm.sig_ratios import get_sig_ratios
from pyPPG.ppg_bm.ppg_derivs import get_ppg_derivs
from pyPPG.ppg_bm.derivs_ratios import get_derivs_ratios
from pyPPG.ppg_bm.statistics import get_statistics

class BmCollection:

    ###########################################################################
    ######################## Initialization of Biomarkers #####################
    ###########################################################################
    def __init__(self, s: pyPPG.PPG, fp: pyPPG.Fiducials):
        """
        The purpose of the Biomarkers class is to calculate the ppg biomarkers.

        :param s: a struct of PPG signal
        :type s: pyPPG.PPG object
        :param fp: a struct of fiducial points
        :type fp: pyPPG.Fiducials object

        """

        self.s = s
        self.fp = fp

    ###########################################################################
    ############################ Get PPG Biomarkers ###########################
    ###########################################################################
    def get_biomarkers (self):
        """
        This function retrieves the list of biomarkers, computes their values, and calculates associated statistics.

        :return:
            - bm_defs: dictionary of biomarkers with name, definition and unit
            - bm_vals: dictionary of biomarkers with values
            - bm_stats: dictionary of biomarkers with statistics
        """

        s=self.s
        fp = self.fp

        ## Get Biomarkers
        bm_ppg_sig, def_ppg_sig = get_ppg_sig(s, fp)
        bm_sig_ratios, def_sig_ratios = get_sig_ratios(s, fp)
        bm_ppg_derivs, def_ppg_derivs = get_ppg_derivs(s, fp)
        bm_derivs_ratios, def_derivs_ratios = get_derivs_ratios(s, fp)

        bm_vals={'ppg_sig': bm_ppg_sig , 'sig_ratios': bm_sig_ratios, 'ppg_derivs': bm_ppg_derivs, 'derivs_ratios': bm_derivs_ratios}
        bm_defs = {'ppg_sig': def_ppg_sig, 'sig_ratios': def_sig_ratios, 'ppg_derivs': def_ppg_derivs, 'derivs_ratios': def_derivs_ratios}

        ## Get Statistics
        bm_stats = get_statistics(fp.sp, fp.on, bm_vals)

        return bm_defs, bm_vals, bm_stats



import pyPPG

import copy
import pandas as pd
import numpy as np
from dotmap import DotMap
from scipy.signal import kaiserord, firwin, filtfilt, detrend, periodogram, lfilter, find_peaks, firls, resample
from scipy import signal

class FpCollection:

    ###########################################################################
    ###################### Initialization of Fiducial Points ##################
    ###########################################################################
    def __init__(self, s: pyPPG.PPG):
        """
        The purpose of the FiducialPoints class is to calculate the fiducial points.

        :param s: object of PPG signal
        :type s: pyPPG.PPG object

        """

        keys=s.__dict__.keys()
        keys_list = list(keys)
        for i in keys_list:
            exec('self.'+i+' = s.'+i)

    ###########################################################################
    ############################ Get Fiducial Points ##########################
    ###########################################################################
    def get_fiducials(self, s: pyPPG.PPG, correct=True):
        '''The function calculates the PPG Fiducial Points.
            - Original signal: List of systolic peak, pulse onset, dicrotic notch, and diastolic peak
            - 1st derivative: List of points of 1st maximum and minimum in 1st derivitive between the onset to onset intervals (u,v)
            - 2nd derivative: List of maximum and minimum points in 2nd derivitive between the onset to onset intervals (a, b, c, d, e)

        :param s: object of PPG signal
        :type s: pyPPG.PPG object
        :param correct: a bool for fiducials points corretion
        :type correct: bool

        :param fiducials: a DataFrame where the key is the name of the fiducial pints and the value is the list of fiducial points PPG Fiducials Points.
        :type fiducials: DataFrame
        '''

        # 'ABD' refers the original Aboy peak detectorm and 'PPGdet' refers the improved version.
        peak_detector='PPGdet'

        # Extract Fiducial Points
        drt0_fp=pd.DataFrame()
        peaks, onsets = self.get_peak_onset(peak_detector)
        dicroticnotch = self.get_dicrotic_notch(peaks, onsets)

        drt1_fp = self.get_1st_drt_fiducials(onsets)
        drt2_fp = self.get_2nd_drt_fiducials(onsets, peaks)
        drt3_fp = self.get_3rd_drt_fiducials(onsets, drt2_fp)

        diastolicpeak = self.get_diastolic_peak(onsets, dicroticnotch, drt2_fp.e)

        # Merge Fiducial Points
        keys=('on', 'sp', 'dn','dp')
        dummy = np.empty(len(peaks))
        dummy.fill(np.NaN)
        n=0
        for temp_val in (onsets,peaks,dicroticnotch,diastolicpeak):
            drt0_fp[keys[n]] = dummy
            drt0_fp[keys[n]][0:len(temp_val)]=temp_val
            n=n+1

        fiducials=pd.DataFrame()
        for temp_drt in (drt0_fp,drt1_fp,drt2_fp,drt3_fp):
            for key in list(temp_drt.keys()):
                fiducials[key] = dummy
                temp_val = temp_drt[key].values
                fiducials[key][0:len(temp_val)]=temp_val

        # Correct Fiducial Points
        if correct:
            fiducials=self.correct_fiducials(fiducials)

        fiducials=fiducials.astype('Int64')
        return fiducials


    ###########################################################################
    ############################ PPG beat detector ############################
    ###########################################################################
    def get_peak_onset(self, peak_detector: str):
        '''PPGdet detects beats in a photoplethysmogram (PPG) signal
        using the improved 'Automatic Beat Detection' of Aboy M et al.

        :param peak_detector: type of peak detector
        :type peak_detector: str

        :return:
            - peaks, 1-d array: indices of detected systolic peaks
            - onsets, 1-d array: indices of detected pulse onsets

        Reference
        ---------
        Aboy M et al., An automatic beat detection algorithm for pressure signals.
        IEEE Trans Biomed Eng 2005; 52: 1662 - 70. <https://doi.org/10.1109/TBME.2005.855725>

        Author:
        Marton A. Goda – Faculty of Biomedical Engineering,
        Technion – Israel Institute of Technology, Haifa, Israel (August 2022)

        Original Matlab implementation:
        Peter H. Charlton – King's College London (August 2017) – University of Cambridge (February 2022)
        <https://github.com/peterhcharlton/ppg-beats>

        Changes from Charlton's implementation:
            1) Detect Maxima:
                *  Systolic peak-to-peak distance is predicted by the heart rate estimate over the preceding 10 sec window.
                *  The peak location is estimated by distances and prominences of the previous peaks.
            2) Find Onsets:
                *  The onset is a local minimum, which is always calculated from the peak that follows it within a given time window
            3) Tidy of Peaks and Onsets:
                *  There is a one-to-one correspondence between onsets and peaks
                *  There are only onset and peak pairs
                *  The distance between the onset and peak pairs can't be smaller than 30 ms
            4) Correct Peaks and Onsets:
                * The Peaks must be the highest amplitude between two consecutive pulse onsets, if not, then these are corrected
                * After the correction of Peaks, the Onsets are recalculated
        '''

        # inputs
        x = copy.deepcopy(self.filt_sig)                    #signal
        fso=self.fs
        fs = 75
        x = resample(x, int(len(self.filt_sig)*(fs/fso)))
        up = self.set_beat_detection()                 #settings
        win_sec=10
        w = fs * win_sec                                    #window length(number of samples)
        win_starts = np.array(list(range(0,len(x),round(0.8*w))))
        win_starts = win_starts[0:min(np.where([win_starts >= len(x) - w])[1])]
        win_starts = np.insert(win_starts,len(win_starts), len(x) + 1 - w)

        # before pre-processing
        hr_win=0  #the estimated systolic peak-to-peak distance, initially it is 0
        hr_win_v=[]
        px = self.detect_maxima(x, 0, hr_win, peak_detector) # detect all maxima
        if len(px)==0:
            peaks = []
            onsets = []
            return peaks, onsets

        # detect peaks in windows
        all_p4 = []
        all_hr = np.empty(len(win_starts)-1)
        all_hr [:] = np.NaN
        hr_past = 0 # the actual heart rate
        hrvi = 0    # heart rate variability index

        for win_no in range(0,len(win_starts) - 1):
            curr_els = range(win_starts[win_no],win_starts[win_no] + w)
            curr_x = x[curr_els]

            y1 = self.def_bandpass(curr_x, fs, 0.9 * up.fl_hz, 3 * up.fh_hz)   # Filter no.1
            hr = self.estimate_HR(y1, fs, up, hr_past)               # Estimate HR from weakly filtered signal
            hr_past=hr
            all_hr[win_no] = hr

            if (peak_detector=='PPGdet') and (hr>40):
                if win_no==0:
                    p1 = self.detect_maxima(y1, 0, hr_win, peak_detector)
                    tr = np.percentile(np.diff(p1), 50)
                    pks_diff = np.diff(p1)
                    pks_diff = pks_diff[pks_diff>=tr]
                    hrvi = np.std(pks_diff) / np.mean(pks_diff) * 5

                hr_win = fs / ((1 + hrvi) * 3)
                hr_win_v.append(hr_win)
            else:
                hr_win=0

            y2 = self.def_bandpass(curr_x, fs, 0.9 * up.fl_hz, 2.5 * hr / 60)           # Filter no. 2
            y2_deriv = self.estimate_deriv(y2)                                          # Estimate derivative from highly filtered signal
            p2 = self.detect_maxima(y2_deriv, up.deriv_threshold,hr_win, peak_detector) # Detect maxima in derivative
            y3 = self.def_bandpass(curr_x, fs, 0.9 * up.fl_hz, 10 * hr / 60)
            p3 = self.detect_maxima(y3, 50, hr_win, peak_detector)                      # Detect maxima in moderately filtered signal
            p4 = self.find_pulse_peaks(p2, p3)
            p4 = np.unique(p4)

            if peak_detector=='PPGdet':
                if len(p4)>round(win_sec/2):
                    pks_diff = np.diff(p4)
                    tr = np.percentile(pks_diff, 30)
                    pks_diff = pks_diff[pks_diff >= tr]

                    med_hr=np.median(all_hr[np.where(all_hr>0)])
                    if ((med_hr*0.5<np.mean(pks_diff)) and (med_hr*1.5<np.mean(pks_diff))):
                        hrvi = np.std(pks_diff) / np.mean(pks_diff)*10

            all_p4 = np.concatenate((all_p4, win_starts[win_no] + p4), axis=None)

        all_p4=all_p4.astype(int)
        all_p4 = np.unique(all_p4)

        peaks, fn = self.correct_IBI(all_p4, px, np.median(all_hr), fs, up)

        peaks = (all_p4/fs*fso).astype(int)
        onsets, peaks = self.find_onsets(self.filt_sig, fso, up, peaks,60/np.median(all_hr)*fs)

        # Correct Peaks
        for i in range(0, len(peaks) - 1):
            max_loc = np.argmax(self.filt_sig[onsets[i]:onsets[i + 1]]) + onsets[i]
            if peaks[i] != max_loc:
                peaks[i] = max_loc

        # Correct onsets
        onsets, peaks = self.find_onsets(self.filt_sig, fso, up, peaks, 60 / np.median(all_hr) * fs)

        temp_i = np.where(np.diff(onsets) == 0)[0]
        if len(temp_i) > 0:
            peaks = np.delete(peaks, temp_i)
            onsets = np.delete(onsets, temp_i)

        temp_i = np.where((peaks - onsets) < fso / 30)[0]
        if len(temp_i) > 0:
            peaks = np.delete(peaks, temp_i)
            onsets = np.delete(onsets, temp_i)

        return peaks, onsets

    ###########################################################################
    ############################# Maximum detector ############################
    ###########################################################################
    def detect_maxima(self, sig: np.array, percentile: int ,hr_win: int, peak_detector: str):
        #Table VI pseudocode
        """
        Detect Maxima function detects all peaks in the raw and also in the filtered signal to find.

        :param sig: array of signal with shape (N,) where N is the length of the signal
        :type sig: 1-d array
        :param percentile: in each signal partition, a rank filter detects the peaks above a given percentile
        :type percentile: int
        :param hr_win: window for adaptive the heart rate estimate
        :type hr_win: int
        :param peak_detector: type of peak detector
        :type peak_detector: str

        :return: maximum peaks of signal, 1-d array.

        """

        tr = np.percentile(sig, percentile)

        if peak_detector=='ABD':

            s1,s2,s3 = sig[2:], sig[1:-1],sig[0:-2]
            m = 1 + np.array(np.where((s1 < s2) & (s3 < s2)))
            max_pks = m[sig[m] > tr]

        if peak_detector=='PPGdet':
            s1,s2,s3 = sig[2:], sig[1:-1],sig[0:-2]

            max_loc = []
            min_loc = []
            max_pks=[]
            intensity_v = []
            if hr_win == 0:
                m = 1 + np.array(np.where((s1 < s2) & (s3 < s2)))
                max_pks = m[sig[m] > tr]
            else:
                max_loc = find_peaks(sig, distance=hr_win)[0]
                min_loc = find_peaks(-sig, distance=hr_win)[0]

                for i in range(0,len(max_loc)):
                    values = abs(max_loc[i] - min_loc)
                    min_v = min(values)
                    min_i = np.where(min_v==values)[0][0]
                    intensity_v.append(sig[max_loc[i]] - sig[min_loc[min_i]])

                # possible improvements:
                #   - using adaptive thresholding for the maximum
                #   - estimate probability density of the maximum

                tr2 = np.mean(intensity_v)*0.25
                max_pks = find_peaks(sig+min(sig),prominence=tr2,distance=hr_win)[0]

        return max_pks

    ###########################################################################
    ############################ Bandpass filtering ###########################
    ###########################################################################
    def def_bandpass(self, sig: np.array, fs: int, lower_cutoff: float, upper_cutoff: float):
        """
        def_bandpass filter function detects all peaks in the raw and also in the filtered signal to find.

        :param sig: array of signal with shape (N,) where N is the length of the signal
        :type sig: 1-d array
        :param fs: sampling frequency
        :type fs: int
        :param lower_cutoff: lower cutoff frequency
        :type lower_cutoff: float
        :param upper_cutoff: upper cutoff frequency
        :type upper_cutoff: float

        :return: bandpass filtered signal, 1-d array.

        """

        # Filter characteristics: Eliminate VLFs (below resp freqs): For 4bpm cutoff
        up = DotMap()
        up.paramSet.elim_vlf.Fpass = 1.3*lower_cutoff   #in Hz
        up.paramSet.elim_vlf.Fstop = 0.8*lower_cutoff   #in Hz
        up.paramSet.elim_vlf.Dpass = 0.05
        up.paramSet.elim_vlf.Dstop = 0.01

        # Filter characteristics: Eliminate VHFs (above frequency content of signals)
        up.paramSet.elim_vhf.Fpass = 1.2*upper_cutoff   #in Hz
        up.paramSet.elim_vhf.Fstop = 0.8*upper_cutoff   #in Hz
        up.paramSet.elim_vhf.Dpass = 0.05
        up.paramSet.elim_vhf.Dstop = 0.03

        # perform BPF
        s = DotMap()
        s.v = sig
        s.fs = fs

        b, a = signal.iirfilter(5, [2 * np.pi * lower_cutoff, 2 * np.pi * upper_cutoff], rs=60,
                                btype='band', analog=True, ftype='cheby2')

        bpf_sig = filtfilt(b, 1, s.v)

        return bpf_sig

    ###########################################################################
    ################### Filter the high frequency components  #################
    ###########################################################################
    def elim_vlfs(self, s: np.array, up: DotMap, lower_cutoff: float):
        """
        This function filter the high frequency components.

        :param s: array of signal with shape (N,) where N is the length of the signal
        :type s: 1-d array
        :param up: setup up parameters of the algorithm
        :type up: DotMap
        :param lower_cutoff: lower cutoff frequency
        :type lower_cutoff: float

        :return: high frequency filtered signal, 1-d array.

        """

        ## Filter pre-processed signal to remove frequencies below resp
        # Adapted from RRest

        ## Eliminate nans
        s.v[np.isnan(s.v)] = np.mean(s.v[~np.isnan(s.v)])

        ##Make filter
        fc=lower_cutoff
        ripple=-20*np.log10(up.paramSet.elim_vlf.Dstop)
        width=abs(up.paramSet.elim_vlf.Fpass-up.paramSet.elim_vlf.Fstop)/(s.fs/2)
        [N,beta] = kaiserord(ripple,width)
        if N * 3 > len(s):
            N = round(N / 3)
        b = firwin(N, fc * 2 / s.fs, window=('kaiser', beta), scale=('True'))
        AMfilter = b

        s_filt=DotMap()
        try:
            s_filt.v = filtfilt(AMfilter, 1, s.v)
            s_filt.v = s.v-s_filt.v
        except:
            s_filt.v = s.v

        s_filt.fs = s.fs

        return s_filt

    ###########################################################################
    ################### Filter the low frequency components  ##################
    ###########################################################################
    def elim_vhfs(self, s: np.array, up: DotMap, upper_cutoff: float):
        """
        This function filter the high frequency components.

        :param s: array of signal with shape (N,) where N is the length of the signal
        :type s: 1-d array
        :param up: setup up parameters of the algorithm
        :type up: DotMap
        :param upper_cutoff: lower cutoff frequency
        :type upper_cutoff: float

        :return: low frequency filtered signal, 1-d array.

        """


        ## Filter signal to remove VHFs
        # Adapted from RRest
        s_filt = DotMap()

        ##Eliminate nans
        s.v[np.isnan(s.v)] = np.mean(s.v[~np.isnan(s.v)])

        ##Check to see if sampling freq is at least twice the freq of interest
        if (up.paramSet.elim_vhf.Fpass/(s.fs/2)) >= 1:
            s_filt.v = s.v
            return

        fc = upper_cutoff
        ripple = -20 * np.log10(up.paramSet.elim_vhf.Dstop)
        width = abs(up.paramSet.elim_vhf.Fpass - up.paramSet.elim_vhf.Fstop) / (s.fs / 2)
        [N, beta] = kaiserord(ripple, width)
        if N * 3 > len(s):
            N = round(N / 3)
        b = firwin(N, fc * 2 / s.fs, window=('kaiser', beta), scale=('True'))
        AMfilter = b

        ## Remove VHFs
        s_dt=detrend(s.v)
        s_filt.v = filtfilt(AMfilter, 1, s_dt)

        return s_filt

    ###########################################################################
    ########################### Heart Rate estimation #########################
    ###########################################################################
    def estimate_HR(self, sig: np.array, fs: int, up: DotMap, hr_past: int):
        """
        Heart Rate Estimation function estimate the heart rate according to the previous heart rate in given time window

        :param sig: array of signal with shape (N,) where N is the length of the signal
        :type sig: 1-d array
        :type fs: int
        :param up: setup up parameters of the algorithm
        :type up: DotMap
        :param hr_past: the average heart rate in the past in given time window
        :type hr_past: int

        :return: estimated heart rate, 1-d array.

        """

        # Estimate PSD
        blackman_window = np.blackman(len(sig))
        f, pxx = periodogram(sig,fs, blackman_window)
        ph = pxx
        fh = f

        # Extract HR
        if (hr_past / 60 < up.fl_hz) | (hr_past / 60 > up.fh_hz):
            rel_els = np.where((fh >= up.fl_hz) & (fh <= up.fh_hz))
        else:
            rel_els = np.where((fh >= hr_past / 60 * 0.5) & (fh <= hr_past / 60 * 1.4))

        rel_p = ph[rel_els]
        rel_f = fh[rel_els]
        max_el = np.where(rel_p==max(rel_p))
        hr = rel_f[max_el]*60
        hr = int(hr[0])

        return hr

    ###########################################################################
    ############# Estimate derivative from highly filtered signal #############
    ###########################################################################
    def estimate_deriv(self, sig: np.array):
        """
        Derivative Estimation function estimate derivative from highly filtered signal based on the
        General least-squares smoothing and differentiation by the convolution (Savitzky Golay) method

        :param sig: array of signal with shape (N,) where N is the length of the signal
        :type sig: 1-d array

        :return: derivative, 1-d array.

        """

        #Savitzky Golay
        deriv_no = 1
        win_size = 5
        deriv = self.savitzky_golay(sig, deriv_no, win_size)

        return deriv


    def savitzky_golay(self, sig: np.array, deriv_no: int, win_size: int):
        """
        This function estimate the Savitzky Golay derivative from highly filtered signal

        :param sig: array of signal with shape (N,) where N is the length of the signal
        :type sig: 1-d array
        :param deriv_no: number of derivative
        :type deriv_no: int
        :param win_size: size of window
        :type win_size: int

        :return: Savitzky Golay derivative, 1-d array.

        """

        ##assign coefficients
        # From: https: // en.wikipedia.org / wiki / Savitzky % E2 % 80 % 93 Golay_filter  # Tables_of_selected_convolution_coefficients
        # which are calculated from: A., Gorry(1990). "General least-squares smoothing and differentiation by the convolution (Savitzky?Golay) method".Analytical Chemistry. 62(6): 570?3. doi: 10.1021 / ac00205a007.

        if deriv_no==0:
            #smoothing
            if win_size == 5:
                coeffs = [-3, 12, 17, 12, -3]
                norm_factor = 35
            elif win_size == 7:
                coeffs = [-2, 3, 6, 7, 6, 3, -2]
                norm_factor = 21
            elif win_size == 9:
                coeffs = [-21, 14, 39, 54, 59, 54, 39, 14, -21]
                norm_factor = 231
            else:
                print('Can''t do this window size')
        elif deriv_no==1:
            # first derivative
            if win_size == 5:
                coeffs = range(-2,3)
                norm_factor = 10
            elif win_size == 7:
                coeffs = range(-3,4)
                norm_factor = 28
            elif win_size == 9:
                coeffs = range(-4,5)
                norm_factor = 60
            else:
                print('Can''t do this window size')
        elif deriv_no == 2:
            # second derivative
            if win_size == 5:
                coeffs = [2, -1, -2, -1, 2]
                norm_factor = 7
            elif win_size == 7:
                coeffs = [5, 0, -3, -4, -3, 0, 5]
                norm_factor = 42
            elif win_size == 9:
                coeffs = [28, 7, -8, -17, -20, -17, -8, 7, 28]
                norm_factor = 462
            else:
                print('Can''t do this window size')
        elif deriv_no == 3:
            # third derivative
            if win_size == 5:
                coeffs = [-1, 2, 0, -2, 1]
                norm_factor = 2
            elif win_size == 7:
                coeffs = [-1, 1, 1, 0, -1, -1, 1]
                norm_factor = 6
            elif win_size == 9:
                coeffs = [-14, 7, 13, 9, 0, -9, -13, -7, 14]
                norm_factor = 198
            else:
                print('Can''t do this window size')
        elif deriv_no == 4:
            # fourth derivative
            if win_size == 7:
                coeffs = [3, -7, 1, 6, 1, -7, 3]
                norm_factor = 11
            elif win_size == 9:
                coeffs = [14, -21, -11, 9, 18, 9, -11, -21, 14]
                norm_factor = 143
            else:
                print('Can''t do this window size')
        else:
            print('Can''t do this order of derivative')


        if deriv_no % 2 == 1:
            coeffs = -np.array(coeffs)

        A = [1, 0]
        filtered_sig = lfilter(coeffs, A, sig)
        # filtered_sig = filtfilt(coeffs, A, sig)
        s = len(sig)
        half_win_size = np.floor(win_size * 0.5)
        zero_pad=filtered_sig[win_size] * np.ones(int(half_win_size))
        sig_in=filtered_sig[win_size-1:s]
        sig_end=filtered_sig[s-1] * np.ones(int(half_win_size))
        deriv = [*zero_pad,*sig_in,*sig_end]
        deriv = deriv / np.array(norm_factor)

        return deriv

    ###########################################################################
    ############################# Pulse detection #############################
    ###########################################################################
    def find_pulse_peaks(self, p2: np.array, p3: np.array):
        """
        Pulse detection function detect the pulse peaks according to the peaks of 1st and 2nd derivatives
        General least-squares smoothing and differentiation by the convolution (Savitzky Golay) method

        :param p2: peaks of the 1st derivatives
        :type p2: 1-d array
        :param p3: peaks of the 2nd derivatives
        :type p2: 1-d array

        :return: pulse peaks, 1-d array.

        """

        p4 = np.empty(len(p2))
        p4[:] = np.NaN
        for k in range(0,len(p2)):
            rel_el = np.where(p3>p2[k])
            if np.any(rel_el) and ~np.isnan(rel_el[0][0]):
                p4[k] = p3[rel_el[0][0]]

        p4 = p4[np.where(~np.isnan(p4))]
        p4 = p4.astype(int)
        return p4

    ###########################################################################
    ####################### Correct peaks' location error #####################
    ###########################################################################
    def  correct_IBI(self, p: np.array, m: np.array, hr: float, fs: int, up: DotMap):
        """
        This function corrects the peaks' location (interbeat intervals) error

        :param p: systolic peaks of the PPG signal
        :type p: 1-d array
        :param m: all maxima of the PPG signal
        :type m: 1-d array
        :param hr: heart rate
        :type hr: float
        :param fs: sampling frequency
        :type fs: int
        :param up: setup up parameters of the algorithm
        :type up: DotMap

        :return: onsets, 1-d array.

        """

        #Correct peaks' location error due to pre-processing
        pc = np.empty(len(p))
        pc[:] = np.NaN
        pc1=[]
        for k in range(0,len(p)):
            temp_pk=abs(m - p[k])
            rel_el = np.where(temp_pk==min(temp_pk))
            pc1=[*pc1,*m[rel_el]]

        # Correct false positives
        # identify FPs
        d = np.diff(pc1)/fs    # interbeat intervals in secs
        fp = self.find_reduced_IBIs(d, hr, up)

        # remove FPs
        pc = np.array(pc1)[fp]

        # Correct false negatives
        d = np.diff(pc)/fs    # interbeat intervals in secs
        fn = self.find_prolonged_IBIs(d, hr, up)

        pc = pc1

        return pc, fn

    def find_reduced_IBIs(self,IBIs: np.array, med_hr: float, up: DotMap):
        """
        This function finds the reduced interbeat intervals

        :param IBIs: interbeat intervals in secs
        :type IBIs: 1-d array
        :param med_hr: median heart rate
        :type med_hr: float
        :param up: setup up parameters of the algorithm
        :type up: DotMap

        :return: fp, the false positive case

        """

        IBI_thresh = up.lower_hr_thresh_prop*60/med_hr
        fp = IBIs < IBI_thresh
        fp = [*np.where(fp == 0)[0].astype(int)]
        return fp

    def find_prolonged_IBIs(self, IBIs: np.array, med_hr: float, up: DotMap):
        """
        This function finds the prolonged interbeat intervals

        :param IBIs: interbeat intervals in secs
        :type IBIs: 1-d array
        :param med_hr: median heart rate
        :type med_hr: float
        :param up: setup up parameters of the algorithm
        :type up: DotMap

        :return: fn, the false negative case

        """
        IBI_thresh = up.upper_hr_thresh_prop*60/med_hr
        fn = IBIs > IBI_thresh

        return fn

    ###########################################################################
    ####################### Setup up the beat detector ########################
    ###########################################################################
    def set_beat_detection(self):
        """
        This function setups the filter parameters of the algorithm

        :return: filter parameters of the algorithm, DotMap

        """
        # plausible HR limits
        up=DotMap()
        up.fl = 30               #lower bound for HR
        up.fh = 200              #upper bound for HR
        up.fl_hz = up.fl/60
        up.fh_hz = up.fh/60

        # Thresholds
        up.deriv_threshold = 75          #originally 90
        up.upper_hr_thresh_prop = 2.25   #originally 1.75
        up.lower_hr_thresh_prop = 0.5    #originally 0.75

        # Other parameters
        up.win_size = 10    #in secs

        return up

    ###########################################################################
    ############################## Find PPG onsets ############################
    ###########################################################################
    def find_onsets(self,sig: np.array, fs: int, up: DotMap, peaks: np.array, med_hr: float):
        """
        This function finds the onsets of PPG sigal

        :param sig: array of signal with shape (N,) where N is the length of the signal
        :type sig: 1-d array
        :param fs: sampling frequency
        :type fs: int
        :param up: setup up parameters of the algorithm
        :type up: DotMap
        :param peaks: peaks of the signal
        :type peaks: 1-d array
        :param med_hr: median heart rate
        :type med_hr: float

        :return: onsets, 1-d array

        """

        Y1=self.def_bandpass(sig, fs, 0.9*up.fl_hz, 3*up.fh_hz)
        temp_oi0=find_peaks(-Y1,distance=med_hr*0.3)[0]

        null_indexes = np.where(temp_oi0<peaks[0])
        if len(null_indexes[0])!=0:
            if len(null_indexes[0])==1:
                onsets = [null_indexes[0][0]]
            else:
                onsets = [null_indexes[0][-1]]
        else:
            onsets = [peaks[0]-round(fs/50)]

        i=1
        while i < len(peaks):
            min_SUT=fs*0.12     # minimum Systolic Upslope Time 120 ms
            min_DT=fs*0.3       # minimum Diastolic Time 300 ms

            before_peak=temp_oi0 <peaks[i]
            after_last_onset=temp_oi0 > onsets[i - 1]
            SUT_time=peaks[i]-temp_oi0>min_SUT
            DT_time = temp_oi0-peaks[i-1]  > min_DT
            temp_oi1 = temp_oi0[np.where(before_peak * after_last_onset*SUT_time*DT_time)]
            if len(temp_oi1)>0:
                if len(temp_oi1) == 1:
                    onsets.append(temp_oi1[0])
                else:
                    onsets.append(temp_oi1[-1])
                i=i+1
            else:
                peaks = np.delete(peaks, i)

        return onsets,peaks

    ###########################################################################
    ########################## Detect dicrotic notch ##########################
    ###########################################################################
    def get_dicrotic_notch (self, peaks: np.array, onsets: list):
        """
        Dicrotic Notch function estimate the location of dicrotic notch in between the systolic and diastolic peak

        :param peaks: peaks of the signal
        :type peaks: 1-d array
        :param onsets: onsets of the signal
        :type onsets: list

        :return: location of dicrotic notches, 1-d array
        """

        ## The 2nd derivative and Hamming low pass filter is calculated.
        dxx = np.diff(np.diff(self.filt_sig))
        fs = self.fs

        # Make filter
        Fn = fs / 2                                 # Nyquist Frequency
        FcU = 20                                    # Cut off Frequency: 20 Hz
        FcD = FcU + 5                               # Transition Frequency: 5 Hz

        n = 21                                      # Filter order
        f = [0, (FcU / Fn), (FcD / Fn), 1]          # Frequency band edges
        a = [1, 1, 0, 0]                            # Amplitudes
        b = firls(n, f, a)

        lp_filt_sig = filtfilt(b, 1,  dxx)          # Low pass filtered signal with 20 cut off Frequency and 5 Hz Transition width

        ## The weighting is calculated and applied to each beat individually
        def t_wmax(i, peaks,onsets):
            if i < 3:
                HR = np.mean(np.diff(peaks))/fs
                t_wmax = -0.1 * HR + 0.45
            else:
                t_wmax = np.mean(peaks[i - 3:i]-onsets[i - 3:i])/fs
            return t_wmax

        dic_not=[]
        for i in range(0,len(onsets)-1):
            nth_beat = lp_filt_sig[onsets[i]:onsets[i + 1]]

            i_Pmax=peaks[i]-onsets[i]
            t_Pmax=(peaks[i]-onsets[i])/fs
            t=np.linspace(0,len(nth_beat)-1,len(nth_beat))/fs
            T_beat=(len(nth_beat)-1)/fs
            tau=(t-t_Pmax)/(T_beat-t_Pmax)
            tau[0:i_Pmax] = 0
            beta=5

            if len(peaks)>1:
                t_w=t_wmax(i, peaks, onsets)
            else:
                t_w=np.NaN

            if t_w!=T_beat:
                tau_wmax=(t_w-t_Pmax)/(T_beat-t_Pmax)
            else:
                tau_wmax=0.9

            alfa=(beta*tau_wmax-2*tau_wmax+1)/(1-tau_wmax)
            if (alfa > 4.5) or (alfa < 1.5):
                HR = np.mean(np.diff(peaks))/fs
                t_w = -0.1 * HR + 0.45
                tau_wmax = (t_w - t_Pmax) / (T_beat - t_Pmax)
                alfa = (beta * tau_wmax - 2 * tau_wmax + 1) / (1 - tau_wmax)

            ## Calculate the Dicrotic Notch for each heart cycle using the weighted window
            if alfa>1:
                w = tau ** (alfa - 1) * (1 - tau) ** (beta - 1)
            else:
                w = tau * (1 - tau) ** (beta - 1)

            pp=w*nth_beat
            pp = pp[np.where(~np.isnan(pp))]
            max_pp_v = np.max(pp)
            max_pp_i=np.where(pp==max_pp_v)[0][0]
            ## NOTE!! Shifting with 26 ms. FIX IT!
            shift=int(self.fs*0.026)
            dic_not.append(max_pp_i+onsets[i]+shift)

        return dic_not

    ###########################################################################
    ########################## Detect diastolic peak ##########################
    ###########################################################################
    def get_diastolic_peak(self, onsets: list, dicroticnotch: list, e_point: pd.Series):
        """
        Dicrotic Notch function estimate the location of dicrotic notch in between the systolic and diastolic peak

        :param onsets: onsets of the signal
        :type onsets: list
        :param dicroticnotches: dicrotic notches of the signal
        :type dicroticnotches: list
        :param e_point: e-points of the signal
        :type e_point: pd.Series

        :return: diastolicpeak location of dicrotic notches, 1-d array
        """

        nan_v = np.empty(len(dicroticnotch))
        nan_v[:] = np.NaN
        diastolicpeak = nan_v

        for i in range(0,len(dicroticnotch)):
            try:
                len_segments=np.diff(onsets)*0.80
                end_segment=int(onsets[i]+len_segments[i])
                try:
                    start_segment = int(dicroticnotch[i])
                    temp_segment = self.filt_sig[start_segment:end_segment]
                    max_locs, _ = find_peaks(temp_segment)

                    if len(max_locs)==0:
                        start_segment = int(e_point[i])
                        temp_segment = self.filt_d1[start_segment:end_segment]
                        max_locs, _ = find_peaks(temp_segment)

                except:
                    pass

                max_dn = max_locs[0] + start_segment
                diastolicpeak[i] = max_dn
            except:
                pass

        return diastolicpeak

    ###########################################################################
    ####################### Get First Derivitive Points #######################
    ###########################################################################
    def get_1st_drt_fiducials(self, onsets: list):
        """Calculate first derivitive points u and v from the PPG' signal

        :param onsets: onsets of the signal
        :type onsets: list

        :return:
            - u: The highest amplitude between the pulse onset and systolic peak on PPG'
            - v: The lowest amplitude between the u-point and diastolic peak on PPG'
            - w: The first local maximum or inflection point after the dicrotic notch on PPG’
        """

        dx = self.filt_d1

        nan_v = np.empty(len(onsets)-1)
        nan_v[:] = np.NaN
        u, v, w = copy.deepcopy(nan_v),copy.deepcopy(nan_v),copy.deepcopy(nan_v),

        for i in range(0,len(onsets)-1):
            try:
                segment = dx[onsets[i]:onsets[i + 1]]

                # u fiducial point
                max_loc = np.argmax(segment)+onsets[i]
                u[i]=max_loc

                # v fiducial point
                upper_bound_coeff = 0.66
                v_upper_bound = ((onsets[i + 1] - onsets[i]) * upper_bound_coeff + onsets[i]).astype(int)
                min_loc = np.argmin(dx[int(u[i]):v_upper_bound])+u[i]-1
                v[i] = min_loc

                # w fiducial point
                temp_segment=dx[int(v[i]):onsets[i+1]]
                max_locs, _ = find_peaks(temp_segment)
                max_w = max_locs[0] + v[i] - 1
                w[i] = max_w

            except:
                pass

        drt1_fp = pd.DataFrame({"u":[], "v":[], "w":[]})
        drt1_fp.u, drt1_fp.v, drt1_fp.w = u, v, w
        return drt1_fp

    ###########################################################################
    ####################### Get Second Derivitive Points ######################
    ###########################################################################
    def get_2nd_drt_fiducials(self, onsets: list, peaks: np.array):
        """Calculate Second derivitive points a, b, c, d, e, and f from the PPG" signal

        :param onsets: onsets of the signal
        :type onsets: list
        :param peaks: peaks of the signal
        :param types: 1-d array

        :return:
            - a: The highest amplitude between pulse onset and systolic peak on PPG"
            - b: The first local minimum after the a-point on PPG"
            - c: The local maximum with the highest amplitude between the b-point and e-point, or if no local maximum is present then the inflection point on PPG"
            - d: The local minimum with the lowest amplitude between the c-point and e-point, or if no local minimum is present then the inflection point on PPG"
            - e: The local maximum with the highest amplitude after the b-point and before the diastolic peak on PPG"
            - f: The first local minimum after the e-point on PPG"
        """

        sig = self.filt_sig
        ddx = self.filt_d2
        dddx = self.filt_d3

        nan_v = np.empty(len(onsets)-1)
        nan_v[:] = np.NaN
        a, b, c, d, e, f = copy.deepcopy(nan_v),copy.deepcopy(nan_v),copy.deepcopy(nan_v),copy.deepcopy(nan_v),copy.deepcopy(nan_v),copy.deepcopy(nan_v)
        for i in range(0,len(onsets)-1):

            try:
                # a fiducial point
                temp_pk=np.argmax(sig[onsets[i]:onsets[i + 1]])+onsets[i]-1
                temp_segment=ddx[onsets[i]:temp_pk]
                max_locs, _ = find_peaks(temp_segment)
                try:
                    max_loc = max_locs[np.argmax(temp_segment[max_locs])]
                except:
                    max_loc = temp_segment.argmax()

                max_a = max_loc + onsets[i] - 1
                a[i] = max_a

                # b fiducial point
                temp_segment=ddx[int(a[i]):onsets[i+1]]
                min_locs, _ = find_peaks(-temp_segment)
                min_b = min_locs[0] + a[i] - 1
                b[i] = min_b

                # e fiducial point
                e_lower_bound = peaks[i]
                upper_bound_coeff = 0.6
                e_upper_bound = ((onsets[i + 1] - onsets[i]) * upper_bound_coeff + onsets[i]).astype(int)
                temp_segment=ddx[int(e_lower_bound):int(e_upper_bound)]
                max_locs, _ = find_peaks(temp_segment)
                if max_locs.size==0:
                    temp_segment=ddx[int(peaks[i]):onsets[i + 1]]
                    max_locs, _ = find_peaks(temp_segment)

                try:
                    max_loc = max_locs[np.argmax(temp_segment[max_locs])]
                except:
                    max_loc = temp_segment.argmax()

                max_e = max_loc + e_lower_bound - 1
                e[i] = max_e

                # c fiducial point
                temp_segment = ddx[int(b[i]):int(e[i])]
                max_locs, _ = find_peaks(temp_segment)
                if max_locs.size>0:
                    max_loc = max_locs[0]
                else:
                    temp_segment = dddx[int(b[i]):int(e[i])]
                    min_locs, _ = find_peaks(-temp_segment)

                    if min_locs.size > 0:
                        max_loc = min_locs[np.argmin(temp_segment[min_locs])]
                    else:
                        max_locs, _ = find_peaks(temp_segment)
                        try:
                            max_loc = max_locs[0]
                        except:
                            max_loc = temp_segment.argmax()

                max_c = max_loc + b[i] - 1
                c[i] = max_c

                # d fiducial point
                temp_segment = ddx[int(c[i]):int(e[i])]
                min_locs, _ = find_peaks(-temp_segment)
                if min_locs.size > 0:
                    min_loc = min_locs[np.argmin(temp_segment[min_locs])]
                    min_d = min_loc + c[i] - 1
                else:
                    min_d = max_c

                d[i] = min_d

                # f fiducial point
                temp_segment = ddx[int(e[i]):onsets[i + 1]]
                min_locs, _ = find_peaks(-temp_segment)
                if (min_locs.size > 0) and (min_locs[0]<len(sig)*0.8):
                    min_loc = min_locs[0]
                else:
                    min_loc = 0

                min_f = min_loc + e[i] - 1
                f[i] = min_f
            except:
                pass

        drt2_fp = pd.DataFrame({"a":[], "b":[], "c":[],"d":[], "e":[], "f":[]})
        drt2_fp.a, drt2_fp.b, drt2_fp.c, drt2_fp.d, drt2_fp.e, drt2_fp.f = a, b, c, d, e, f
        return drt2_fp

    def get_3rd_drt_fiducials(self, onsets: list, drt2_fp: pd.DataFrame):
        """Calculate third derivitive points p1 and p2 from the PPG'" signal

            :param onsets: onsets of the signal
            :type onsets: list
            :param drt2_fp: fiducial points of PPG" signal
            :type drt2_fp: DataFrame

            :return:
                - p1: The first local maximum after the b-point on PPG'"
                - p2: The last local minimum after the b-point and before the d-point on PPG'"

        """
        dddx = self.filt_d3

        nan_v = np.empty(len(onsets)-1)
        nan_v[:] = np.NaN
        p1, p2 = copy.deepcopy(nan_v), copy.deepcopy(nan_v)

        for i in range(0, len(onsets) - 1):
            try:
                # p1 fiducial point
                ref_b = drt2_fp.b[np.squeeze(np.where(np.logical_and(drt2_fp.b > onsets[i], drt2_fp.b < onsets[i + 1])))]
                if ref_b.size == 0:
                    ref_b = onsets[i]

                temp_segment = dddx[int(ref_b):onsets[i + 1]]
                max_locs, _ = find_peaks(temp_segment)
                try:
                    max_loc = max_locs[0]
                except:
                    max_loc = temp_segment.argmax()

                max_p1 = max_loc + ref_b - 1
                p1[i] = max_p1

                # p2 fiducial point
                ref_start = p1[i]
                ref_c = drt2_fp.c[np.squeeze(np.where(np.logical_and(drt2_fp.c > onsets[i], drt2_fp.c < onsets[i + 1])))]
                ref_d = drt2_fp.d[np.squeeze(np.where(np.logical_and(drt2_fp.d > onsets[i], drt2_fp.d < onsets[i + 1])))]

                if ref_d > ref_c:
                    ref_end = ref_d
                    min_ind = -1
                elif ref_c.size > 0:
                    ref_start = ref_c
                    ref_end = onsets[i + 1]
                    min_ind = 0
                elif drt2_fp.e.size > 0:
                    ref_end = onsets[i + 1]
                    min_ind = 0

                temp_segment = dddx[int(ref_start):int(ref_end)]
                min_locs, _ = find_peaks(-temp_segment)
                if min_locs.size > 0:
                    min_p2 = min_locs[min_ind] + ref_start - 1
                else:
                    min_p2 = ref_end

                p2[i] = min_p2

            except:
                pass

        drt3_fp = pd.DataFrame({"p1": [], "p2": []})
        drt3_fp.p1, drt3_fp.p2 = p1, p2

        return drt3_fp

    def correct_fiducials(self,fiducials: pd.DataFrame):
        """Correct the Fiducial Points

            :param fiducials: a dictionary where the key is the name of the fiducial pints and the value is the list of fiducial points.
            :type fiducials: DataFrame

            :return:
                - fiducials: a dictionary where the key is the name of the fiducial pints and the value is the list of fiducial points
        """

        for i in range(0,len(fiducials.on)):

            # Correct onset
            try:
                win_onr = self.fs * 0.005
                if fiducials.on[i]>win_onr:
                    win_onl = win_onr
                else:
                    win_onl = fiducials.on[i]

                min_loc = np.argmax(self.filt_sig[fiducials.on[i]+win_onl:fiducials.on[i]+win_onr]) + fiducials.on[i]
                if fiducials.on[i] != min_loc:

                    if fiducials.a[i] > self.fs*0.075:
                        win_a = int(self.fs*0.075)
                    else:
                        win_a = int(fiducials.a[i])

                    fiducials.on[i] = np.argmax(self.filt_d3[int(fiducials.a[i]) - win_a:int(fiducials.a[i])]) + int(fiducials.a[i]) - win_a
            except:
                pass

            # Correct dicrotic notch
            try:
                temp_segment=self.filt_sig[int(fiducials.sp[i]):int(fiducials.dp[i])]
                min_dn=find_peaks(-temp_segment)[0]+fiducials.sp[i]
                diff_dn=abs(min_dn-fiducials.dp[i])
                if len(min_dn)>0 and diff_dn>round(self.fs/100):
                    fiducials.dn[i]=min_dn
                    try:
                        strt_dn = int(fiducials.sp[i])
                        stp_dn = int(fiducials.f[i])
                        fiducials.dn[i] = find_peaks(-self.filt_sig[strt_dn:stp_dn])[0][-1] + strt_dn
                        if fiducials.dn[i] > min_dn:
                            fiducials.dn[i] = min_dn
                    except:
                        strt_dn = fiducials.e[i]
                        stp_dn = fiducials.f[i]
                        fiducials.dn[i] = np.argmin(self.filt_d3[strt_dn:stp_dn])+strt_dn
                        if fiducials.dn[i] > min_dn:
                            fiducials.dn[i] = min_dn

            except:
                pass

            # Correct v and w-point
            try:
                if fiducials.v[i]> fiducials.e[i]:
                    fiducials.v[i] = np.argmin(self.filt_d1[int(fiducials.u[i]):int(fiducials.e[i])]) + int(fiducials.u[i])
                    fiducials.w[i] = find_peaks(self.filt_d1[int(fiducials.v[i]):int(fiducials.f[i])])[0][0] + int(fiducials.v[i])
            except:
                pass

            # Correct w-point
            try:
                temp_end=int(np.diff(fiducials.on[i:i+2])*0.8)
                temp_segment=self.filt_d1[int(fiducials.dn[i]):int(fiducials.on[i]+temp_end)]
                min_w=find_peaks(-temp_segment)[0]+fiducials.dn[i]

                if fiducials.w[i] < fiducials.e[i]:
                    fiducials.w[i] = np.argmax(self.filt_d1[int(fiducials.e[i]):int(fiducials.f[i])]) + int(fiducials.e[i])

                # if fiducials.w[i] > fiducials.f[i]:
                #     fiducials.w[i] = int(fiducials.f[i])

                if fiducials.w[i] > min_w:
                    fiducials.w[i] = min_w

            except:
                pass

            # Correct f-point
            try:
                temp_end=int(np.diff(fiducials.on[i:i+2])*0.8)
                temp_segment=self.filt_d2[int(fiducials.e[i]):int(fiducials.on[i]+temp_end)]
                min_f=np.argmin(temp_segment)+fiducials.e[i]

                if fiducials.w[i] > fiducials.f[i]:
                    fiducials.f[i] = min_f
            except:
                pass


        # Correct diastolic peak
        try:
            fiducials.dp = self.get_diastolic_peak(fiducials.on, fiducials.dn, fiducials.e)
        except:
            pass

        return fiducials


import pyPPG

import pandas as pd
from pyPPG.ppg_bm.bm_extraction import get_biomarkers

###########################################################################
####################### Get Biomarkers of PPG Signal ######################
###########################################################################
def get_ppg_sig(s: pyPPG.PPG, fp: pyPPG.Fiducials):
    """
    This function returns the biomarkers of PPG signal.

    :param s: object of PPG signal
    :type s: pyPPG.PPG object
    :param fp: object of fiducial points
    :type fp: pyPPG.Fiducials object

    :return:
        - biomarkers: dictionary of biomarkers of PPG signal
        - biomarkers_lst: list a biomarkers with name, definition and unit
    """

    biomarkers_lst = [
                    ["Tpi",   "Pulse interval, the time between the pulse onset and pulse offset", "[s]"],
                    ["Tpp",   "Peak-to-peak interval, the time between two consecutive systolic peaks", "[s]"],
                    ["Tsys",  "Systolic time, the time between the pulse onset and dicrotic notch", "[s]"],
                    ["Tdia",  "Diastolic time, the time is between the dicrotic notch and pulse offset", "[s]"],
                    ["Tsp",   "Systolic peak time, the time between the pulse onset and systolic peak", "[s]"],
                    ["Tdp",	  "Diastolic peak time, the time between the pulse onset and diastolic peak", "[s]"],
                    ["deltaT","Time delay, the time between the systolic peak and diastolic peak", "[s]"],
                    ["Tsw10", "Systolic width, the width at 10% of the systolic peak amplitude between the pulse onset and systolic peak", "[s]"],
                    ["Tsw25", "Systolic width, the width at 25% of the systolic peak amplitude between the pulse onset and systolic peak", "[s]"],
                    ["Tsw33", "Systolic width, the width at 33% of the systolic peak amplitude between the pulse onset and systolic peak", "[s]"],
                    ["Tsw50", "Systolic width, the width at 50% of the systolic peak amplitude between the pulse onset and systolic peak", "[s]"],
                    ["Tsw66", "Systolic width, the width at 66% of the systolic peak amplitude between the pulse onset and systolic peak", "[s]"],
                    ["Tsw75", "Systolic width, the width at 75% of the systolic peak amplitude between the pulse onset and systolic peak", "[s]"],
                    ["Tsw90", "Systolic width, the width at 90% of the systolic peak amplitude between the pulse onset and systolic peak", "[s]"],
                    ["Tdw10", "Diastolic width, the width at 10% of the systolic peak amplitude between the systolic peak and pulse offset", "[s]"],
                    ["Tdw25", "Diastolic width, the width at 25% of the systolic peak amplitude between the systolic peak and pulse offset", "[s]"],
                    ["Tdw33", "Diastolic width, the width at 33% of the systolic peak amplitude between the systolic peak and pulse offset", "[s]"],
                    ["Tdw50", "Diastolic width, the width at 50% of the systolic peak amplitude between the systolic peak and pulse offset", "[s]"],
                    ["Tdw66", "Diastolic width, the width at 66% of the systolic peak amplitude between the systolic peak and pulse offset", "[s]"],
                    ["Tdw75", "Diastolic width, the width at 75% of the systolic peak amplitude between the systolic peak and pulse offset", "[s]"],
                    ["Tdw90", "Diastolic width, the width at 90% of the systolic peak amplitude between the systolic peak and pulse offset", "[s]"],
                    ["Tpw10", "Pulse width, the sum of the systolic width and the diastolic width at 10%", "[s]"],
                    ["Tpw25", "Pulse width, the sum of the systolic width and the diastolic width at 25%", "[s]"],
                    ["Tpw33", "Pulse width, the sum of the systolic width and the diastolic width at 33%", "[s]"],
                    ["Tpw50", "Pulse width, the sum of the systolic width and the diastolic width at 50%", "[s]"],
                    ["Tpw66", "Pulse width, the sum of the systolic width and the diastolic width at 66%", "[s]"],
                    ["Tpw75", "Pulse width, the sum of the systolic width and the diastolic width at 75%", "[s]"],
                    ["Tpw90", "Pulse width, the sum of the systolic width and the diastolic width at 90%", "[s]"],
                    ["Asp",   "Systolic peak amplitude, the difference in amplitude between the pulse onset and systolic peak", "[nu]"],
                    ["Adn",   "Dicrotic notch amplitude, the difference in amplitude between the pulse onset and dicrotic notch", "[nu]"],
                    ["Adp",   "Diastolic peak amplitude, the difference in amplitude between the pulse onset and diastolic peak", "[nu]"],
                    ["Aoff",  "Pulse onset amplitude, the difference in amplitude between the pulse onset and pulse offset", "[nu]"],
                    ["AUCpi", "Area under pulse interval curve, the area under the pulse wave between pulse onset and pulse offset", "[nu]"],
                    ["AUCsys","Area under systolic curve, the area under the pulse wave between the pulse onset and the dicrotic notch", "[nu]"],
                    ["AUCdia","Area under diastolic curve, the area under the pulse wave between the dicrotic notch and pulse offset", "[nu]"],
    ]

    header = ['name', 'definition', 'unit']
    biomarkers_lst = pd.DataFrame(biomarkers_lst, columns=header)

    df, df_biomarkers = get_biomarkers(s, fp, biomarkers_lst.name)

    return df_biomarkers, biomarkers_lst



import pyPPG

import numpy as np
from dotmap import DotMap
import pandas as pd
from scipy.signal import find_peaks

###########################################################################
####################### PPG biomarkers extraction #########################
###########################################################################

class BmExctator:
    """
    Class that extracts the PPG biomarkers.
    """

    def __init__(self, data: DotMap, peak_value: float, peak_time: float, next_peak_value: float, next_peak_time: float, onsets_values: np.array, onsets_times: np.array,
                 sample_rate: int, list_biomarkers: list, fiducials: pd.DataFrame):
        """

        :param data: struct of PPG,PPG', PPG", PPG'":

            - data.sig: segment of PPG timeseries to analyse and extract biomarkers as a np array
            - data.d1: segment of PPG'
            - data.d2: segment of PPG"
            - data.d3: segment of PPG'"
        :type data: DotMap
        :param peak_value: PPG peak value
        :type peak_value: float
        :param peak_time: the time corresponding to the peak detected
        :type peak_time: float
        :param next_peak_value: PPG next peak value
        :type next_peak_value: float
        :param next_peak_time: the time corresponding to the peak detected
        :type next_peak_time: float
        :param onsets_values: array of PPG two onsets values surrounding the peak
        :type onsets_values: Numpy array
        :param onsets_times: array of the two times corresponding to each onset detected
        :type onsets_times: Numpy array
        :param sample_rate: segment data sample rate
        :type sample_rate: int
        :param list_biomarkers: list of biomarkers
        :type list_biomarkers: list
        :param fiducials: location of fiducial points of the given pulse wave
        :type fiducials: DataFrame

        """

        self.fiducials=fiducials
        self.list_biomarkers=list_biomarkers
        self.segment = data.sig
        self.segment_d1 = data.d1
        self.segment_d2 = data.d2
        self.segment_d3 = data.d3
        self.peak_value = peak_value
        self.peak_time = peak_time
        self.next_peak_value = next_peak_value
        self.next_peak_time = next_peak_time
        self.onsets_values = onsets_values
        self.onsets_times = onsets_times
        self.sample_rate = sample_rate
        self.dn, self.dp, self.Tdn, self.Tdp = self._getDicroticNotchDiastolicPeak()
        self.u, self.v, self.w, self.Tu, self.Tv, self.Tw = self._getFirstDerivitivePoints()
        self.a, self.b, self.c, self.d, self.e, self.f, self.Ta, self.Tb, self.Tc, self.Td, self.Te, self.Tf = self._getSecondDerivitivePoints()
        self.p1, self.p2, self.Tp1, self.Tp2 = self._getThirdDerivitivePoints()

    def map_func(self):
        """ This function assign for each name of biomarkers a function that calculates it

        :returns my_funcs: a dictionary where the key is the name of the biomarker and the value is the function to call
        """
        my_funcs = {#PPG signal
                    "Tpi": self.getTpi(),
                    "Tpp": self.getTpp(),
                    "Tsys": self.getTsys(),
                    "Tdia": self.getTdia(),
                    "Tsp": self.getTsp(),
                    "Tdp": self.getTdp(),
                    "deltaT": self.get_deltaT(),
                    "Tsw10": self.getSystolicWidth_d_percent(10),
                    "Tsw25": self.getSystolicWidth_d_percent(25),
                    "Tsw33": self.getSystolicWidth_d_percent(33),
                    "Tsw50": self.getSystolicWidth_d_percent(50),
                    "Tsw66": self.getSystolicWidth_d_percent(66),
                    "Tsw75": self.getSystolicWidth_d_percent(75),
                    "Tsw90": self.getSystolicWidth_d_percent(90),
                    "Tdw10": self.getDiastolicWidth_d_percent(10),
                    "Tdw25": self.getDiastolicWidth_d_percent(25),
                    "Tdw33": self.getDiastolicWidth_d_percent(33),
                    "Tdw50": self.getDiastolicWidth_d_percent(50),
                    "Tdw66": self.getDiastolicWidth_d_percent(66),
                    "Tdw75": self.getDiastolicWidth_d_percent(75),
                    "Tdw90": self.getDiastolicWidth_d_percent(90),
                    "Tpw10": self.getSumSW_DW(10),
                    "Tpw25": self.getSumSW_DW(25),
                    "Tpw33": self.getSumSW_DW(33),
                    "Tpw50": self.getSumSW_DW(50),
                    "Tpw66": self.getSumSW_DW(66),
                    "Tpw75": self.getSumSW_DW(75),
                    "Tpw90": self.getSumSW_DW(90),
                    "Asp": self.getSystolicPeak(),
                    "Adn": self.getDicroticNotchAmplitude(),
                    "Adp": self.getDiastolicPeak(),
                    "Aoff": self.getPulseOffsetAmplitude(),
                    "AUCpi": self.getAUCpi(),
                    "AUCsys": self.getAUCsys(),
                    "AUCdia": self.getAUCdia(),

                    # Signal ratios
                    "IPR": self.getIPR(),
                    "Tsys/Tdia": self.get_ratio_Tsys_Tdia(),
                    "Tpw25/Tpi": self.get_ratio_Tpwx_Tpi(25),
                    "Tpw50/Tpi": self.get_ratio_Tpwx_Tpi(50),
                    "Tpw75/Tpi": self.get_ratio_Tpwx_Tpi(75),
                    "Tpw25/Tsp": self.get_ratio_Tpwx_Tsp(25),
                    "Tpw50/Tsp": self.get_ratio_Tpwx_Tsp(50),
                    "Tpw75/Tsp": self.get_ratio_Tpwx_Tsp(75),
                    "Tdw10/Tsw10": self.get_ratio_Tdwx_Tswx(10),
                    "Tdw25/Tsw25": self.get_ratio_Tdwx_Tswx(25),
                    "Tdw33/Tsw33": self.get_ratio_Tdwx_Tswx(33),
                    "Tdw50/Tsw50": self.get_ratio_Tdwx_Tswx(50),
                    "Tdw66/Tsw66": self.get_ratio_Tdwx_Tswx(66),
                    "Tdw75/Tsw75": self.get_ratio_Tdwx_Tswx(75),
                    "Tdw90/Tsw90": self.get_ratio_Tdwx_Tswx(90),
                    "Tsp/Tpi": self.get_ratio_Tsp_Tpi(),
                    "Asp/Aoff": self.get_ratio_Asp_Aoff(),
                    "Adp/Asp": self.get_ratio_Adp_Asp(),
                    "IPA": self.getIPA(),
                    "Tsp/Asp": self.get_ratio_Tsp_Asp(),
                    "Asp/deltaT": self.get_ratio_Asp_deltaT(),
                    "Asp/(Tpi-Tsp)": self.get_ratio_Asp_TpiTsp(),

                    # PPG derivatives
                    "u": self.get_u(),
                    "v": self.get_v(),
                    "w": self.get_w(),
                    "a": self.get_a(),
                    "b": self.get_b(),
                    "c": self.get_c(),
                    "d": self.get_d(),
                    "e": self.get_e(),
                    "f": self.get_f(),
                    "Tu": self.get_Tu(),
                    "Tv": self.get_Tv(),
                    "Tw": self.get_Tw(),
                    "Ta": self.get_Ta(),
                    "Tb": self.get_Tb(),
                    "Tc": self.get_Tc(),
                    "Td": self.get_Td(),
                    "Te": self.get_Te(),
                    "Tf": self.get_Tf(),
                    "Tb-c": self.get_Tbc(),
                    "Tb-d": self.get_Tbd(),
                    "Tp1": self.get_Tp1(),
                    "Tp2": self.get_Tp2(),
                    "Tp1-dp": self.get_Tp1_dp(),
                    "Tp2-dp": self.get_Tp2_dp(),

                    # Derivatives ratios
                    "Tu/Tpi": self.get_ratio_Tu_Tpi(),
                    "Tv/Tpi": self.get_ratio_Tv_Tpi(),
                    "Tw/Tpi": self.get_ratio_Tw_Tpi(),
                    "Ta/Tpi": self.get_ratio_Ta_Tpi(),
                    "Tb/Tpi": self.get_ratio_Tb_Tpi(),
                    "Tc/Tpi": self.get_ratio_Tc_Tpi(),
                    "Td/Tpi": self.get_ratio_Td_Tpi(),
                    "Te/Tpi": self.get_ratio_Te_Tpi(),
                    "Tf/Tpi": self.get_ratio_Tf_Tpi(),
                    "(Tu-Ta)/Tpi": self.get_ratio_TuTa_Tpi(),
                    "(Tv-Tb)/Tpi": self.get_ratio_TvTb_Tpi(),
                    "Au/Asp":self.get_ratio_Au_Asp(),
                    "Av/Au":self.get_ratio_Av_Au(),
                    "Aw/Au":self.get_ratio_Aw_Au(),
                    "Ab/Aa":self.get_ratio_Ab_Aa(),
                    "Ac/Aa":self.get_ratio_Ac_Aa(),
                    "Ad/Aa":self.get_ratio_Ad_Aa(),
                    "Ae/Aa":self.get_ratio_Ae_Aa(),
                    "Af/Aa":self.get_ratio_Af_Aa(),
                    "Ap2/Ap1": self.get_ratio_Ap2_Ap1(),
                    "(Ac-Ab)/Aa":self.get_ratio_AcAb_Aa(),
                    "(Ad-Ab)/Aa":self.get_ratio_AdAb_Aa(),
                    "AGI":self.getAGI(),
                    "AGImod":self.getAGImod(),
                    "AGIinf":self.getAGIinf(),
                    "AI": self.getAI(),
                    "RIp1": self.getRIp2(),
                    "RIp2": self.getRIp2(),
                    "SC": self.getSC(),
                    "IPAD": self.getIPAD(),
        }
        return my_funcs

    def get_biomarker_extract_func(self):
        """ This function go through the list of biomarkers and call the function that is relevant to calculate it each biomarker takes two spots in the biomarker vector: one for the average value and one for its standard deviation.

            :returns biomarkers_vec: a vector of each biomarker avg value and std of the patient"""
        my_funcs = self.map_func()
        biomarkers_vec = []

        for biomarker in self.list_biomarkers:
            func_to_call = my_funcs[biomarker]
            biomarkers_vec.append(func_to_call)
        return biomarkers_vec

    def _getPeaksOnsets(self,x):
        """Find the peaks and onsets of a short FILTERED segment of PPG

        :return: peaks
        :return: onsets
        """
        peaks, _ = find_peaks(x)
        onsets, _ = find_peaks(-x)
        return peaks, onsets

    def _getDicroticNotchDiastolicPeak(self):
        """Calculate Dicrotic Notch and Diastolic Peak of PPG

        :return: dn: Sample distance from PPG onset to the Dicrotic Notch on PPG
        :return: dp: Sample distance from PPG onset to the Diastolic Peak on PPG
        :return: Tdn: Time from PPG onset to the Dicrotic Notch on PPG
        :return: Tdp: Time from PPG onset to the Diastolic Peak on PPG
        """

        dn = (self.fiducials.dn-self.fiducials.on).values[0]
        dp = (self.fiducials.dp-self.fiducials.on).values[0]
        Tdn = dn / self.sample_rate
        Tdp = dp / self.sample_rate

        return dn, dp, Tdn, Tdp
    def _getFirstDerivitivePoints(self):
        """Calculate first derivitive points from a SINGLE Onset-Onset segment of PPG'

        :return: u: Sample distance from PPG onset to the greatest maximum peak between the left systolic onset and the systolic peak on PPG'
        :return: v: Sample distance from PPG onset to the lowest minimum pits between the systolic peak and the right systolic onset on PPG'
        :return: w: Sample distance from PPG onset to the first maximum peak after v on PPG'
        :return: Tu: Time from PPG onset to the greatest maximum peak between the left systolic onset and the systolic peak on PPG'
        :return: Tv: Time from PPG onset to the lowest minimum pits between the systolic peak and the right systolic onset on PPG'
        :return: Tw: Time from PPG onset to the first maximum peak after v on PPG'
        """

        u = (self.fiducials.u-self.fiducials.on).values[0]
        v = (self.fiducials.v-self.fiducials.on).values[0]
        w = (self.fiducials.w - self.fiducials.on).values[0]
        Tu = u / self.sample_rate
        Tv = v / self.sample_rate
        Tw = w / self.sample_rate

        return u, v, w, Tu, Tv, Tw

    def _getSecondDerivitivePoints(self):
        """Calculate second derivitive points from a SINGLE Onset-Onset segment of PPG"

        :return: a: Sample distance from PPG onset to the first maximum peak between left systolic onset and systolic peak on PPG"
        :return: b: Sample distance from PPG onset to the first minimum pits after a on PPG"
        :return: c: Sample distance from PPG onset to the greatest maximum peak between b and e, or if no maximum peak is present then the inflection point on PPG"
        :return: d: Sample distance from PPG onset to the lowest minimum pits between c and e, or if no minimum pits is present then the inflection point on PPG"
        :return: e: Sample distance from PPG onset to the greatest maximum peak between the systolic peak and  the right systolic onset on PPG"
        :return: f: Sample distance from PPG onset to the first minimum pits after e on PPG"
        :return: Ta: Time from PPG onset to the first maximum peak between left systolic onset and systolic peak on PPG"
        :return: Tb: Time from PPG onset to the first minimum pits after a on PPG"
        :return: Tc: Time from PPG onset to the greatest maximum peak between b and e, or if no maximum peak is present then the inflection point on PPG"
        :return: Td: Time from PPG onset to the lowest minimum pits between c and e, or if no minimum pits is present then the inflection point on PPG"
        :return: Te: Time from PPG onset to the greatest maximum peak between the systolic peak and  the right systolic onset on PPG"
        :return: Tf: Time from PPG onset to the first minimum pits after e on PPG"
        """

        a = (self.fiducials.a-self.fiducials.on).values[0]
        b = (self.fiducials.b-self.fiducials.on).values[0]
        c = (self.fiducials.c-self.fiducials.on).values[0]
        d = (self.fiducials.d-self.fiducials.on).values[0]
        e = (self.fiducials.e-self.fiducials.on).values[0]
        f = (self.fiducials.f-self.fiducials.on).values[0]
        Ta = a / self.sample_rate
        Tb = b / self.sample_rate
        Tc = c / self.sample_rate
        Td = d / self.sample_rate
        Te = e / self.sample_rate
        Tf = f / self.sample_rate

        return a, b, c, d, e, f, Ta, Tb, Tc, Td, Te, Tf

    def _getThirdDerivitivePoints(self):
        """Calculate third derivitive points from a SINGLE Onset-Onset segment of PPG'"

        :return: p1: Sample distance from PPG onset to the first local maximum after b on PPG'"
        :return: p2: Sample distance from PPG onset to the last local minimum before d, if c = d, then the first local minimum after d on PPG'"
        :return: Tp1: Time from PPG onset to the first local maximum after b on PPG'"
        :return: Tp2: Time from PPG onset to last local minimum before d, if c = d, then the first local minimum after d on PPG'"
        """

        p1 = (self.fiducials.p1-self.fiducials.on).values[0]
        p2 = (self.fiducials.p2-self.fiducials.on).values[0]
        Tp1 = p1 / self.sample_rate
        Tp2 = p2 / self.sample_rate

        return p1, p2, Tp1, Tp2
    def _find_nearest(self, arr, value):
        """ This function calculates the index in an array of the closest value to the arg value

            :param arr: the array where to find the index
            :param value: the value to be compared with

            :return: idx: the index of the value closest to the arg value
            """
        idx = (np.abs(arr - value)).argmin()
        return idx

    def _getTime(self, vec, val):
        """ get the time of a value in the PPG waveform  data vector
            :param vec: the array where to find the index
            :param val: the value to be compared with
            :return: index: the index of the value closest to the arg value
            """
        tmp_vec = np.array([vec[i]-val for i in range(0, len(vec))])
        index = self._find_nearest(tmp_vec, 0)
        return index

    def _getSysTime_from_val(self, val):
        """ get the time of a value in the PPG waveform  data vector

            :param val: the value from which we need the time in the timeserie

            :return: t_data: the time corresponding to the value
            """
        idx_ons = int(self.onsets_times[0]*self.sample_rate)
        idx_peak = int(self.peak_time*self.sample_rate)
        idx = idx_peak - idx_ons
        vec_data = np.array(self.segment[0: idx])
        t_val = self._getTime(vec_data, val)
        t_data = t_val + idx_ons
        return t_data

    def _getDiaTime_from_val(self, val):
        """ get the time of a value in the PPG waveform  data vector

            :param val: the value from which we need the time in the timeserie

            :return: t_data: the time corresponding to the value
            """
        idx_ons_right = int(self.onsets_times[1]*self.sample_rate)
        idx_ons_left = int(self.onsets_times[0] * self.sample_rate)
        idx_peak = int(self.peak_time*self.sample_rate)
        idx = idx_peak - idx_ons_left
        idx2 = idx_ons_right - idx_ons_left
        vec_data = np.array(self.segment[idx: idx2])
        t_val = self._getTime(vec_data, val)
        t_data = t_val + idx_peak
        return t_data

    def _getBaselineSlope(self):
        """ get the baseline slope in the PPG waveform  data vector

            :return: baseline slope
        """
        left_onset_time = self.onsets_times[0]
        right_onset_time = self.onsets_times[1]
        left_onset_value = self.onsets_values[0]
        right_onset_value = self.onsets_values[1]
        slope_numerator = right_onset_value - left_onset_value
        slope_denom = right_onset_time*self.sample_rate - left_onset_time*self.sample_rate
        return slope_numerator/slope_denom

    def _getBaselineCst(self):
        """ get the difference between the right and the left onset values

            :return: cst
        """
        left_onset_value = self.onsets_values[0]
        right_onset_value = self.onsets_values[1]
        cst = right_onset_value - left_onset_value
        return cst

    def getTpi(self):
        """ Tpi which means the Pulse Interval, the time between the pulse onset and pulse offset.

            :return: Tpi biomarker
        """
        return self.onsets_times[1] - self.onsets_times[0]
    def getTpp(self):
        """ Tpp means the Peak-to-Peak Interval, the time between two consecutive systolic peaks.

            :return:  Tpp biomarker
        """
        cardiac_period = self.next_peak_time - self.peak_time
        return cardiac_period

    def getTsys(self):
        """ Tsys means the Systolic Time, the time between the pulse onset and dicrotic notch.

        :return: Tsys biomarker
        """

        Tsys = self.Tdn
        return Tsys

    def getTdia(self):
        """ Tdia means the Diastolic Time, the time between the dicrotic notch and pulse offset.

        :return: Tdia biomarker
        """

        Tdia = self.getTpi()-self.Tdp
        return Tdia

    def getTsp(self):
        """ Tsp means the Systolic Peak Time, the time between the pulse onset and systolic peak.

        :return: Tsp biomarker
        """
        left_onset = self.onsets_times[0]
        Tsp = self.peak_time - left_onset
        return Tsp

    def getTdp(self):
        """ Tdp means the Diastolic Peak Time, the time between the pulse onset and diastolic peak.

        :return: Tdp biomarker
        """

        Tdp = self.Tdp
        return Tdp

    def get_deltaT(self):
        """ deltaT means the Time Delay, the time between the systolic peak and diastolic peak.

        :return: deltaT biomarker
        """

        deltaT = self.Tdp-self.getTsp()
        return deltaT

    def getSystolicWidth_d_percent(self, d):
        """ The function calculates the Systolic Width,
            the width at x% of the Systolic Peak Amplitude between the pulse onset and systolic peak.

            :param d: the percentage chosen to calculate the width

            :return: Tswx biomarker
        """
        # value in segment corresponding to d percent of pulse height
        d_percent_val = (d/100)*(self.peak_value - self.onsets_values[0]) + self.onsets_values[0]
        time_of_d = self._getSysTime_from_val(d_percent_val)
        Tswx = self.peak_time - (time_of_d/self.sample_rate)
        return Tswx


    def getDiastolicWidth_d_percent(self, d):
        """ The function calculates the Diastolic Width,
            the width at x% of the Systolic Peak Amplitude between the systolic peak and pulse offset.

            :param d: the percentage chosen to calculate the width

            :return: Tdwx biomarker
        """
        # value in segment corresponding to d percent of pulse height
        d_percent_val = (d/100)*(self.peak_value - self.onsets_values[1]) + self.onsets_values[1]
        time_of_d = self._getDiaTime_from_val(d_percent_val)
        Tdwx = (time_of_d/self.sample_rate) - self.peak_time
        return Tdwx

    def getSumSW_DW(self, d):
        """ The function calculates Pulse Width,
            the sum of the Systolic Width and Diastolic Width at x%.

            :param d: the percentage chosen to calculate the width

            :return: Tpwx biomarker
        """
        Tswx = self.getSystolicWidth_d_percent(d)
        Tdwx = self.getDiastolicWidth_d_percent(d)
        return Tswx + Tdwx

    def getSystolicPeak(self):
        """ The function calculates the Systolic Peak Amplitude, the difference in amplitude between the pulse onset and systolic peak.

            :return: Systolic Peak Amplitude biomarker
        """
        sys_peak = self.peak_value - self.onsets_values[0]
        return sys_peak
    def getDicroticNotchAmplitude(self):
        """ The function calculates the Dicrotic Notch Amplitude, the difference in amplitude between the pulse onset and the dicrotic notch.

            :return: Dicrotic Notch Amplitude biomarker
        """
        dn_value = self.segment[self.dn]
        dn_amp = dn_value - self.onsets_values[0]
        return dn_amp

    def getDiastolicPeak(self):
        """ The function calculates the Diastolic Peak Amplitude, the difference in amplitude between the pulse onset and the diastolic peak.

            :return: Diastolic Peak Amplitude biomarker
        """
        ## temp solution 04/04/2023 --> if DN==DP
        dp_value = self.segment[self.dp+20]
        dia_peak = dp_value - self.onsets_values[0]
        return dia_peak

    def getPulseOffsetAmplitude(self):
        """ The function calculates the Pulse Offset Amplitude, the difference in amplitude between the pulse onset and pulse offset.

            :return: Pulse Offset Amplitude biomarker
        """

        offset_value = self.onsets_values[1]
        Aoff = offset_value - self.onsets_values[0]
        return Aoff

    def getAUCpi(self):
        """ The function the Area Under Pulse Interval Curve, the area under the pulse wave between pulse onset and pulse offset.

            :return: AUCpi biomarker
        """
        left_onset_time = self.onsets_times[0]*self.sample_rate
        right_onset_time = self.onsets_times[1]*self.sample_rate
        baseline_shift_slope = self._getBaselineSlope()
        baseline_cst = self._getBaselineCst()
        vec_value_between_ons = self.segment
        num_t = len(vec_value_between_ons)
        baseline = baseline_shift_slope*self.peak_time*self.sample_rate + baseline_cst
        sum = 0
        for t in range(0, num_t):
            sum += vec_value_between_ons[t] - baseline_shift_slope*((t+left_onset_time)) + baseline_cst
        AUCpi = 10*sum/((self.peak_value - baseline) * (right_onset_time - left_onset_time))

        return AUCpi

    def getAUCsys(self):
        """ The function calculates the Area Under Systolic Curve, the area under the pulse wave between the pulse onset and dicrotic notch.

            :return: AUCsys biomarker
        """
        left_onset_time = self.onsets_times[0]*self.sample_rate
        right_onset_time = self.onsets_times[1]*self.sample_rate
        baseline_shift_slope = self._getBaselineSlope()
        baseline_cst = self._getBaselineCst()
        vec_value_between_ons = self.segment
        num_t = self.dn
        baseline = baseline_shift_slope*self.peak_time*self.sample_rate + baseline_cst
        sum = 0
        for t in range(0, num_t):
            sum += vec_value_between_ons[t] - baseline_shift_slope*((t+left_onset_time)) + baseline_cst
        AUCsys = 10*sum/((self.peak_value - baseline) * (right_onset_time - left_onset_time))

        return AUCsys

    def getAUCdia(self):
        """ The function calculates Area Under Diastolic Curve, the area under the pulse wave between the dicrotic notch and pulse offset.

            :return: AUCdia biomarker
        """
        AUCdia = self.getAUCpi()-self.getAUCsys()
        return AUCdia

    def getIPR(self):
        """ The function calculates the Instantaneous Pulse Rate, 60/CP.

            :return: IPR biomarker
        """
        IPR = 60/self.getTpp()
        return IPR

    def get_ratio_Tsys_Tdia(self):
        """ The function calculates the ratio of the Systolic Time to the Diastolic Time.

            :return: Tsys/Tdia biomarker
        """
        Tsys_Tdia = self.getTsys()/self.getTdia()
        return Tsys_Tdia

    def get_ratio_Tpwx_Tpi(self, d):
        """ The function calculates the ratio of the Pulse Width at x% of the Systolic Peak Amplitude to the Systolic Peak Time.

            :param d: the percentage chosen to calculate the width

            :return: The ratio of the Tpi to the Pulse Width biomarker
        """
        sys_width = self.getSystolicWidth_d_percent(d)
        dia_width = self.getDiastolicWidth_d_percent(d)
        width = sys_width + dia_width
        Tpi = self.getTpi()
        return width/Tpi

    def get_ratio_Tpwx_Tsp(self, d):
        """ The function calculates the ratio of the Pulse Width at x% of the Systolic Peak Amplitude to the Systolic Peak Time.

            :param d: the percentage chosen to calculate the width

            :return: Tpwx/Tsp biomarker
        """
        Tswx = self.getSystolicWidth_d_percent(d)
        Tdwx = self.getDiastolicWidth_d_percent(d)
        Tpwx = Tswx + Tdwx
        Tsp = self.getSystolicPeakTime()
        return Tpwx/Tsp

    def get_ratio_Tdwx_Tswx(self, d):
        """ The function calculates the ratio of the Diastolic Width to the Systolic Width at x% width.

            :param d: the percentage chosen to calculate the width

            :return: Tdwx/Tswx biomarker
        """
        Tswx = self.getSystolicWidth_d_percent(d)
        Tdwx = self.getDiastolicWidth_d_percent(d)
        return Tdwx/Tswx

    def get_ratio_Tsp_Tpi(self):
        """ The function calculates the ratio of the Systolic Peak Time to the Pulse Interval

            :return: Tsp/Tpi biomarker
        """
        Tsp = self.getSystolicPeakTime()
        Tpi = self.getTpi()
        return Tsp/Tpi

    def get_ratio_Asp_Aoff(self):
        """ The function calculates the ratio of the Systolic Peak Time to the Pulse Interval

            :return: Asp/Aoff biomarker
        """
        Asp = self.getSystolicPeak()
        Aoff = self.onsets_values[1]
        return Asp/Aoff

    def get_ratio_Adp_Asp(self):
        """ The function calculates Reflection Index, the ratio of the Diastolic Peak Amplitude to the Systolic Peak Amplitude.

            :return: Reflection Index biomarker
        """
        RI = self.getDiastolicPeak()/self.getSystolicPeak()
        return RI
    def getIPA(self):
        """ The function calculates the Inflection Point Area, the ratio of the Area Under Diastolic Curve to the Area Under Systolic Curve.

            :return: IPA biomarker
        """
        IPA = self.getAUCdia()/self.getAUCsys()
        return IPA

    def get_ratio_Tsp_Asp(self):
        """ The function calculates the ratio of the Systolic Peak Time to the Systolic Peak Amplitude.

            :return: Tsp/Asp biomarker
        """

        return self.getTsp()/self.getSystolicPeak()
    def get_ratio_Asp_deltaT(self):
        """ The function calculates the Stiffness Index, the ratio of the Systolic Peak Amplitude to the Time Delay.

            :return: Stiffness Index biomarker
        """
        SI = self.getSystolicPeak()/(self.getTdp()-self.getTsp())
        return SI

    def get_ratio_Asp_TpiTsp(self):
        """ The function calculates the ratio of the Systolic Peak Amplitude to the difference between the Pulse Interval and Systolic Peak Time.

            :return: Asp/(Tpi-Tsp)
        """
        Tpi = self.getTpi()
        Tsp = self.getSystolicPeakTime()
        Asp = self.getSystolicPeak()
        return Asp/(Tpi-Tsp)


    def get_u(self):
        """ u means the u-point sample, the sample between the pulse onset and u-point.

            :return: u biomarker
        """
        return self.u


    def get_v(self):
        """ v means the v-point sample, the sample between the pulse onset and v-point.

            :return: v biomarker
        """
        return self.v

    def get_w(self):
        """ w means the w-point sample, the sample between the pulse onset and w-point.

            :return: w biomarker
        """
        return self.w

    def get_a(self):
        """ a means the a-point sample, the sample between the pulse onset and a-point.

            :return: a biomarker
        """
        return self.a

    def get_b(self):
        """ b means the b-point sample, the sample between the pulse onset and b-point.

            :return: b biomarker
        """
        return self.b

    def get_c(self):
        """ c means the c-point sample, the sample between the pulse onset and c-point.

            :return: c biomarker
        """
        return self.c

    def get_d(self):
        """ d means the d-point sample, the sample between the pulse onset and d-point.

            :return: d biomarker
        """
        return self.d

    def get_e(self):
        """ e means the e-point sample, the sample between the pulse onset and e-point.

            :return: e biomarker
        """
        return self.e

    def get_f(self):
        """ f means the f-point sample, the sample between the pulse onset and f-point.

            :return: f biomarker
        """
        return self.f

    def get_Tu(self):
        """ Tu means the u-point time, the time between the pulse onset and u-point.

            :return: Tu biomarker
        """
        return self.Tu

    def get_Tv(self):
        """ Tv means the v-point time, the time between the pulse onset and v-point.

            :return: Tv biomarker
        """
        return self.Tv

    def get_Tw(self):
        """ Tw means the v-point time, the time between the pulse onset and w-point.

            :return: Tw biomarker
        """
        return self.Tw

    def get_Ta(self):
        """ Ta means the a-point time, the time between the pulse onset and a-point.

            :return: Ta biomarker
        """
        return self.Ta

    def get_Tb(self):
        """ Tb means the b-point time, the time between the pulse onset and b-point.

            :return: Tb biomarker
        """
        return self.Tb

    def get_Tc(self):
        """ Tc means the c-point time, the time between the pulse onset and c-point.

            :return: Tb biomarker
        """
        return self.Tc

    def get_Td(self):
        """ Td means the d-point time, the time between the pulse onset and d-point.

            :return: Td biomarker
        """
        return self.Td


    def get_Te(self):
        """ Te means the e-point time, the time between the pulse onset and e-point.

            :return: Te biomarker
        """
        return self.Te

    def get_Tf(self):
        """ Tf means the f-point time, the time between the pulse onset and f-point.

            :return: Tf biomarker
        """
        return self.Tf

    def get_Tbc(self):
        """ Tbc means the b–c time, the time between the b-point and c-point.

            :return: Tbc biomarker
        """
        return self.Tc-self.Tb

    def get_Tbd(self):
        """ Tbd means the b–d time, the time between the b-point and d-point.

            :return: Tbd biomarker
        """
        return self.Td - self.Tb

    def get_Tp1(self):
        """ Tp1 means the p1-point time, the time between the pulse onset and p1-point.

            :return: Tp1 biomarker
        """
        return self.Tp1

    def get_Tp2(self):
        """ Tp2 means the p1-point time, the time between the pulse onset and p2-point.

            :return: Tp2 biomarker
        """
        return self.Tp2

    def get_Tp1_dp(self):
        """ The function calculatesthe p1–dia time, the time between the p1-point and diastolic peak.

            :return: Tdia-Tp1 biomarker
        """
        Tp1_dia=(self.dp-self.p1)/self.sample_rate
        return Tp1_dia

    def get_Tp2_dp(self):
        """ The function calculatesthe p2–dia time, the time between the p2-point and diastolic peak.

            :return: Tdia-Tp2 biomarker
        """
        Tp2_dia=(self.dp-self.p2)/self.sample_rate
        return Tp2_dia


    def get_ratio_Tu_Tpi(self):
        """ The function calculates the ratio of the u-point time to the Pulse Interval.

            :return: Tu/Tpi biomarker
        """
        T1 = self.get_Tu()
        return T1 / self.getTpp()


    def get_ratio_Tv_Tpi(self):
        """ The function calculates the ratio of the v-point time to the Pulse Interval.

            :return: Tv/Tpi biomarker
        """
        Tv = self.get_Tv()
        return Tv / self.getTpp()

    def get_ratio_Tw_Tpi(self):
        """ The function calculates the ratio of the w-point time to the Pulse Interval.

            :return: Tw/Tpi biomarker
        """
        Tw = self.get_Tw()
        return Tw / self.getTpp()

    def get_ratio_Ta_Tpi(self):
        """ The function calculates the ratio of the a-point time to the Pulse Interval.

            :return: Ta/Tpi biomarker
        """
        Ta = self.get_Ta()
        return Ta / self.getTpp()

    def get_ratio_Tb_Tpi(self):
        """ The function calculates the ratio of the b-point time to the Pulse Interval.

            :return: Tb/Tpi biomarker
        """
        Tb = self.get_Tb()
        return Tb / self.getTpp()

    def get_ratio_Tc_Tpi(self):
        """ The function calculates the ratio of the c-point time to the Pulse Interval.

            :return: Tc/Tpi biomarker
        """
        Tc = self.get_Tc()
        return Tc / self.getTpp()

    def get_ratio_Td_Tpi(self):
        """ The function calculates the ratio of the d-point time to the Pulse Interval.

            :return: Td/Tpi biomarker
        """
        Td = self.get_Td()
        return Td / self.getTpp()

    def get_ratio_Te_Tpi(self):
        """ The function calculates the ratio of the e-point time to the Pulse Interval.

            :return: Te/Tpi biomarker
        """
        Te = self.get_Te()
        return Te / self.getTpp()

    def get_ratio_Tf_Tpi(self):
        """ The function calculates the ratio of the f-point time to the Pulse Interval.

            :return: Tf/Tpi biomarker
        """
        Tf = self.get_Tf()
        return Tf / self.getTpp()

    def get_ratio_TuTa_Tpi(self):
        """ The function calculates the ratio of the difference between the u-point time and a-point time to the Pulse Interval.

            :return: (Tu-Ta)/Tpi biomarker
        """
        Ta = self.get_Ta()
        Tu = self.get_Tu()
        Tpi = self.getTpi()
        return (Tu - Ta) / Tpi

    def get_ratio_TvTb_Tpi(self):
        """ The function calculates the ratio of the difference between the v-point time and b-point time to the Pulse Interval.

            :return: (Tv-Tb)/Tpi
        """
        Tb = self.get_Tb()
        Tv = self.get_Tv()
        Tpi = self.getTpi()
        return (Tv-Tb)/Tpi

    def get_ratio_Au_Asp(self):
        """ This function calculates the ratio of the u-point amplitude to the Systolic Peak Amplitude.

            :return: Au/Asp biomarker
        """
        u_max = self.segment_d1[self.get_u()]
        sp_amp = self.peak_value
        return u_max / sp_amp

    def get_ratio_Av_Au(self):
        """ This function calculates the ratio of the v-point amplitude to the u-point amplitude.

            :return: Av/Au biomarker
        """
        u_max = self.segment_d1[self.get_u()]
        v_min = self.segment_d1[self.get_v()]
        return v_min / u_max

    def get_ratio_Aw_Au(self):
        """ This function calculates the ratio of the w-point amplitude to the u-point amplitude.

            :return: Aw/Au biomarker
        """
        u_max = self.segment_d1[self.get_u()]
        w_max = self.segment_d1[self.get_w()]
        return w_max / u_max

    def get_ratio_Ab_Aa(self):
        """ This function calculates the ratio of the b-point amplitude to the a-point amplitude.

            :return: Ab/Aa biomarker
        """
        a_max = self.segment_d2[self.get_a()]
        b_min = self.segment_d2[self.get_b()]
        return b_min / a_max

    def get_ratio_Ac_Aa(self):
        """ This function calculates the ratio of the c-point amplitude to the a-point amplitude.

            :return: Ac/Aa biomarker
        """
        a_max = self.segment_d2[self.get_a()]
        c_max = self.segment_d2[self.get_c()]
        return c_max / a_max

    def get_ratio_Ad_Aa(self):
        """ This function calculates the ratio of the d-point amplitude to the a-point amplitude.

            :return: Ad/Aa biomarker
        """
        a_max = self.segment_d2[self.get_a()]
        d_min = self.segment_d2[self.get_d()]
        return d_min / a_max

    def get_ratio_Ae_Aa(self):
        """ This function calculates the ratio of the e-point amplitude to the a-point amplitude.

            :return: Ae/Aa biomarker
        """
        a_max = self.segment_d2[self.get_a()]
        e_max = self.segment_d2[self.get_e()]
        return e_max / a_max

    def get_ratio_Af_Aa(self):
        """ This function calculates the ratio of the f-point amplitude to the a-point amplitude.

            :return: Af/Aa biomarker
        """
        a_max = self.segment_d2[self.get_a()]
        f_min = self.segment_d2[self.get_f()]
        return f_min / a_max

    def get_ratio_Ap2_Ap1(self):
        """ The function calculates the ratio of the p2-point amplitude to the p1-point amplitude.

            :return: Ap2/Ap1 biomarker
        """
        Rp2p1 = self.segment[self.p2]/self.segment[self.p1]
        return Rp2p1

    def get_ratio_AcAb_Aa(self):
        """ The function calculates the ratio of the difference between the b-point amplitude and c-point amplitude to the a-point amplitude.

            :return: (Ac-Ab)/Aa biomarker
        """
        Aa = self.segment_d2[self.get_a()]
        Ab = self.segment_d2[self.get_b()]
        Ac = self.segment_d2[self.get_c()]
        return (Ac-Ab)/Aa

    def get_ratio_AdAb_Aa(self):
        """ The function calculates the ratio of the difference between the b-point amplitude and d-point amplitude to the a-point amplitude.

            :return: (Ad-Ab)/Aa biomarker
        """
        Aa = self.segment_d2[self.get_a()]
        Ab = self.segment_d2[self.get_b()]
        Ad = self.segment_d2[self.get_d()]
        return (Ad-Ab)/Aa

    def getAGI(self):
        """ The function calculates the Aging Index.

            :return: (Ab-Ac-Ad-Ae)/Aa biomarker
        """
        Aa = self.segment_d2[self.get_a()]
        Ab = self.segment_d2[self.get_b()]
        Ac = self.segment_d2[self.get_c()]
        Ad = self.segment_d2[self.get_d()]
        Ae = self.segment_d2[self.get_e()]
        return (Ab-Ac-Ad-Ae)/Aa

    def getAGImod(self):
        """ The function calculates the Modified Aging Index.

            :return: (Ab-Ac-Ad)/Aa biomarker
        """
        Aa = self.segment_d2[self.get_a()]
        Ab = self.segment_d2[self.get_b()]
        Ac = self.segment_d2[self.get_c()]
        Ad = self.segment_d2[self.get_d()]
        return (Ab-Ac-Ad)/Aa

    def getAGIinf(self):
        """ The function calculates the Informal Aging Index.

            :return: (Ab-Ae)/Aa biomarker
        """
        Aa = self.segment_d2[self.get_a()]
        Ab = self.segment_d2[self.get_b()]
        Ae = self.segment_d2[self.get_e()]
        return (Ab-Ae)/Aa

    def getAI(self):
        """ The function calculates the Augmentation Index, (PPG(Tp2) − PPG(Tp1))/Asp.

            :return: AI biomarker
        """
        AI = (self.segment[self.p2]-self.segment[self.p1])/self.peak_value
        return AI

    def getRIp1(self):
        """ The function calculates the Reflection Index of p1, Adp/(PPG(Tp1) − PPG(Tpi(0))).

            :return: RIp1 biomarker
        """
        RIp1 = self.getDiastolicPeak()/self.segment[self.p1]
        return RIp1

    def getRIp2(self):
        """ The function calculates the Reflection Index of p2, Adp/(PPG(Tp2) − PPG(Tpi(0))).

            :return: RIp2 biomarker
        """
        RIp2 = self.getDiastolicPeak()/self.segment[self.p2]
        return RIp2

    def getSC(self):
        """ The function calculates the Spring Constant, PPG"(Tsp)/((Asp-Au)/Asp).

            :return: SC biomarker
        """
        ddxSPA=self.segment_d2[(self.getTsp()*self.sample_rate).astype(int)]
        SPA=self.getSystolicPeak()
        MS=self.segment[self.u]
        SC = ddxSPA/((SPA-MS)/SPA)
        return SC

    def getIPAD(self):
        """ The function calculates the Inflection point area plus normalised d-point amplitude, AUCdia/AUCsys+Ad/Aa.

            :return: IPAD biomarker
        """
        IPAD = self.getAUCdia()/self.getAUCsys()+self.get_ratio_Ad_Aa()
        return IPAD
    def getRatioSW_DW(self, d):
        """ The function calculates the ratio of systolic and diastolic width at d percent of the pulse height

            :param d: the percentage chosen to calculate the width

            :return: ratio biomarker
        """
        sw_d = self.getSystolicWidth_d_percent(d)
        dw_d = self.getDiastolicWidth_d_percent(d)
        ratio = dw_d/sw_d
        return ratio

    def getPIR(self):
        """ The function calculates the ratio between the peak value and the right onset value

            :return: pir biomarker
        """
        pir = self.peak_value/self.onsets_values[1]
        return pir

    def getMS(self):
        """ The function calculates Maximum slope, PPG'(u)/(PPG(systolic peak) − PPG(systolic onset))

            :return: MS biomarker
        """
        MS = self.segment_d1[self.u]
        return MS

    def getUpslope(self):
        """ The function calculates Systolic Upslope between the left onset and the systolic peak.

            :return: Systolic Upslope
        """
        left_onset_time = self.onsets_times[0]*self.sample_rate
        left_onset_value = self.onsets_values[0]
        slope_numer = self.peak_value-left_onset_value
        slope_denom = self.peak_time*self.sample_rate - left_onset_time
        return slope_numer/slope_denom

    def getdiffVal(self):
        """ The function calculates the time between the left onset and the systolic peak.

            :return: left onset and systolic peak time
        """
        left_onset_value = self.onsets_values[0]
        diff = self.peak_value - left_onset_value
        return diff

    def getSTT(self):
        """ STT means slope transit time, which based on geometrical considerations of the PPG pulse wave to account for simultaneous.

            :return: STT biomarker
        """
        upslope = self.getUpslope()
        A = self.getdiffVal()
        return A/upslope

    def getSystolicPeakTime(self):
        """ Systolic Peak Time means the distance between the consecutive Systolic Peaks

        :return: Systolic Peak Times
         """
        return self.peak_time - self.onsets_times[0]

    def getSystolicPeakOutputCurve(self):
        """Peak time divided by systolic amplitude

        :return: sys_peak_time/sys_amplitude
        """
        sys_peak_time = self.getSystolicPeakTime()
        sys_amplitude = self.getSystolicPeak()
        return sys_peak_time/sys_amplitude

    def getAGIext(self):
        """ The function calculates the Extended Aging Index.

            :return: (Ab-Ac-Ad-Ae-Af)/Aa biomarker
        """
        Aa = self.segment_d2[self.get_a()]
        Ab = self.segment_d2[self.get_b()]
        Ac = self.segment_d2[self.get_c()]
        Ad = self.segment_d2[self.get_d()]
        Ae = self.segment_d2[self.get_e()]
        Af = self.segment_d2[self.get_e()]
        return (Ab-Ac-Ad-Ae-Af)/Aa

###########################################################################
############################ Get PPG biomarkers ###########################
###########################################################################
def get_biomarkers(s: pyPPG.PPG, fp: pyPPG.Fiducials, biomarkers_lst):
    """
    The function calculates the biomedical biomarkers of PPG signal.

    :param s: object of PPG signal
    :type s: pyPPG.PPG object
    :param fp: object of fiducial points
    :type fp: pyPPG.Fiducials object

    :return:
        - df: data frame with onsets, offset and peaks
        - df_biomarkers: data frame with PPG signal biomarkers
    """

    fs=s.fs
    ppg=s.filt_sig
    data = DotMap()

    df = pd.DataFrame()
    df_biomarkers = pd.DataFrame(columns=biomarkers_lst)
    peaks = fp.sp.values
    onsets = fp.on.values

    for i in range(len(onsets) - 1):
        onset = onsets[i]
        offset = onsets[i + 1]
        data.sig = ppg[int(onset):int(offset)]
        data.d1 = s.filt_d1[int(onset):int(offset)]
        data.d2 = s.filt_d2[int(onset):int(offset)]
        data.d3 = s.filt_d3[int(onset):int(offset)]
        peak = peaks[(peaks > onset) * (peaks < offset)]
        if len(peak) != 1:
            continue
        peak = peak[0]

        temp_fiducials = fp.get_row(i)

        peak_value = ppg[peak]
        peak_time = peak / fs
        onset_value = ppg[onset]
        onset_time = onset / fs

        if (peak_value - onset_value) == 0:
            continue

        offset_value = ppg[offset]
        offset_time = offset / fs

        idx_array = np.where(peaks == peak)
        idx = idx_array[0]
        onsets_values = np.array([onset_value, offset_value])
        onsets_times = np.array([onset_time, offset_time])
        if (idx + 1) < len(peaks):
            next_peak_value = ppg[peaks[idx + 1].astype('int64')][0]
            next_peak_time = peaks[idx + 1] / fs
            next_peak_time = next_peak_time[0]
            try:
                biomarkers_extractor = BmExctator(data, peak_value, peak_time, next_peak_value, next_peak_time, onsets_values, onsets_times, fs, biomarkers_lst,temp_fiducials)
                biomarkers_vec = biomarkers_extractor.get_biomarker_extract_func()
                lst = list(biomarkers_vec)
                df_biomarkers.loc[len(df_biomarkers.index)] = lst
                df = pd.concat({'onset': onset, 'offset': offset, 'peak': peak}, ignore_index=True)
            except:
                pass
        else:
            print("no more peaks")
    return df, df_biomarkers


import numpy as np


def _check_shape_(signal, fs):
    assert len(signal) > fs*5, "Signal must be at list five seconds"
    signal = np.array(signal)
    assert len(signal.shape) > 0, "Signal must not be empty"
    assert len(signal.shape) < 3, "Signal can be 1-dimensional array or a matrix with different signal in every column "


def _check_fragment_PRSA_(d):
    assert d > 0, "The parameter d should be strictly positive"


class WrongParameter(Exception):
    pass


import pyPPG

import pandas as pd
from pyPPG.ppg_bm.bm_extraction import get_biomarkers

###########################################################################
####################### Get Ratios of PPG Derivatives #####################
###########################################################################
def get_derivs_ratios(s: pyPPG.PPG, fp: pyPPG.Fiducials):
    """
    This function returns the biomarkers of Derivatives ratios.

    :param s: object of PPG signal
    :type s: pyPPG.PPG object
    :param fp: object of fiducial points
    :type fp: pyPPG.Fiducials object

    :return:
        - df_biomarkers: dictionary of biomarkers of Derivatives ratios
        - biomarkers_lst: list a biomarkers with name, definition and unit
    """

    biomarkers_lst = [
                    ["Tu/Tpi",       "Ratio of the u-point time vs. the pulse interval", "[%]"],
                    ["Tv/Tpi",       "Ratio of the v-point time vs. the pulse interval", "[%]"],
                    ["Tw/Tpi",       "Ratio of the w-point time vs. the pulse interval", "[%]"],
                    ["Ta/Tpi",       "Ratio of the a-point time vs. the pulse interval", "[%]"],
                    ["Tb/Tpi",       "Ratio of the b-point time vs. the pulse interval", "[%]"],
                    ["Tc/Tpi",       "Ratio of the c-point time vs. the pulse interval", "[%]"],
                    ["Td/Tpi",       "Ratio of the d-point time vs. the pulse interval", "[%]"],
                    ["Te/Tpi",       "Ratio of the e-point time vs. the pulse interval", "[%]"],
                    ["Tf/Tpi",       "Ratio of the f-point time vs. the pulse interval", "[%]"],
                    ["(Tu-Ta)/Tpi",  "Ratio of the difference between the u-point time and a-point time vs. the pulse interval", "[%]"],
                    ["(Tv-Tb)/Tpi",  "Ratio of the difference between the v-point time and b-point time vs. the pulse interval", "[%]"],
                    ["Au/Asp",       "Ratio of the u-point amplitude vs. the systolic peak amplitude", "[%]"],
                    ["Av/Au",        "Ratio of the v-point amplitude vs. the u-point amplitude", "[%]"],
                    ["Aw/Au",        "Ratio of the w-point amplitude vs. the u-point amplitude", "[%]"],
                    ["Ab/Aa",        "Ratio of the b-point amplitude vs. the a-point amplitude", "[%]"],
                    ["Ac/Aa",        "Ratio of the c-point amplitude vs. the a-point amplitude", "[%]"],
                    ["Ad/Aa",        "Ratio of the d-point amplitude vs. the a-point amplitude", "[%]"],
                    ["Ae/Aa",        "Ratio of the e-point amplitude vs. the a-point amplitude", "[%]"],
                    ["Af/Aa",        "Ratio of the f-point amplitude vs. the a-point amplitude", "[%]"],
                    ["Ap2/Ap1",      "Ratio of the p2-point amplitude vs. the p1-point amplitude", "[%]"],
                    ["(Ac-Ab)/Aa",   "Ratio of the difference between the b-point amplitude and c-point amplitude vs. the a-point amplitude", "[%]"],
                    ["(Ad-Ab)/Aa",   "Ratio of the difference between the b-point amplitude and d-point amplitude vs. the a-point amplitude", "[%]"],
                    ["AGI",          "Aging Index, (Ab-Ac-Ad-Ae)/Aa", "[%]"],
                    ["AGImod",       "Modified aging index, (Ab-Ac-Ad)/Aa", "[%]"],
                    ["AGIinf",       "Informal aging index, (Ab-Ae)/Aa", "[%]"],
                    ["AI",           "Augmentation index, (PPG(Tp2)-PPG(Tp1))/Asp", "[%]"],
                    ["RIp1",         "Reflection index of p1, Adp/(PPG(Tp1)-PPG(Tpi(0)))", "[%]"],
                    ["RIp2",         "Reflection index of p2, Adp/(PPG(p2)-PPG(Tpi(0)))", "[%]"],
                    ["SC",           "Spring constant, PPG''(Tsp)/((Asp-Au)/Asp)", "[nu]"],
                    ["IPAD",         "Inflection point area plus normalised d-point amplitude, AUCdia/AUCsys+Ad/Aa", "[nu]"],
    ]

    header = ['name', 'definition', 'unit']
    biomarkers_lst = pd.DataFrame(biomarkers_lst, columns=header)

    df, df_biomarkers = get_biomarkers(s, fp, biomarkers_lst.name)

    return df_biomarkers, biomarkers_lst


from pyPPG import PPG, Fiducials, Biomarkers
from pyPPG.datahandling import load_data, plot_fiducials, save_data
import pyPPG.fiducials as FP
import pyPPG.biomarkers as BM
import pyPPG.ppg_sqi as SQI

import numpy as np
import sys
import json

###########################################################################
################################## EXAMPLE ################################
###########################################################################
def ppg_example(data_path="",start = 0, end = 0, filtering=True, correct=True, savingfolder="temp_dir", savefig=True, savedata=True, savingformat="csv"):
    '''
    This is an example code for PPG analysis. The main parts:
        1) Loading a raw PPG signal: various file formats such as .mat, .csv, .txt, or .edf.
        2) Get Fiducial points: extract the fiducial points of PPG, PPG', PPG'' and PPG'" signals
        3) Plot Fiducials Points
        4) Get Biomarkers: extract 74 PPG biomarkers in four categories:
            - PPG signal
            - Signal ratios
            - PPG derivatives
            - Derivatives ratios
        5) Get Statistics: summary of the 74 PPG biomarkers
        6) Save data: save the extracted Fiducial points, Biomarkers, and Statistics into .csv file

    :param data_path: path of the PPG signal
    :type data_path: str
    :param start: beginning the of signal in sample
    :type start: int
    :param end: end of the signal in sample
    :type end: int
    :param filtering: a bool for filtering
    :type filtering: bool
    :param correct: a bool for fiducials points corretion
    :type correct: bool
    :param savingfolder: location of the saved data
    :type savingfolder: str
    :param savefig: a bool for current figure saving
    :type savefig: bool
    :param savedata: a bool for saving fiducial points, biomarkers, and statistics
    :type savedata: bool
    :param savingformat: file format of the saved date, the provided file formats .mat and .csv
    :type savingformat: str

    :return: fiducial points, a dictionary where the key is the name of the fiducial pints and the value is the list of fiducial points


    Example:

        .. code-block:: python

            from pyPPG.example import ppg_example

            # run example code
            ppg_example(savedata=True, savefig=True)

    '''

    ## Loading a raw PPG signal
    ppg_data = load_data(data_path,start,end,filtering)
    s = PPG(ppg_data)

    ## Get Fiducial points
    fpex = FP.FpCollection(s)
    fiducials=fpex.get_fiducials(s,correct)+s.start
    fp = Fiducials(fiducials)

    if savefig:
        ## Plot Fiducials Points
        plot_fiducials(s, fp, savingfolder)

    if savedata:
        ## Get Biomarkers and Statistics
        bmex = BM.BmCollection(s, fp)
        bm_defs, bm_vals, bm_stats = bmex.get_biomarkers()
        bm = Biomarkers(bm_defs, bm_vals , bm_stats)

        ## Save data
        save_data(s, fp, bm, savingformat, savingfolder)

    # PPG SQI
    ppgSQI=round(np.mean(SQI.get_ppgSQI(s.filt_sig, s.fs, fp.sp))*100,2)
    print('Mean PPG SQI: ', ppgSQI, '%')

    print('Program finished')
    return fiducials


###########################################################################
############################## RUN EXAMPLE CODE ###########################
###########################################################################
if __name__ == "__main__":

    if len(sys.argv) > 1:
        input_data = json.loads(sys.argv[1])
        function_name = input_data['function']
        function_args = input_data['args']

        if function_name == 'ppg_example':
            fiducials = ppg_example(**function_args)
            json_data = fiducials.to_json(orient="split")
            print(json.dumps(json_data))
        else:
            print("Invalid function name")
    else:
        print("Please provide function name and arguments as JSON string")
        ppg_example(savedata=True, savefig=True)


import pyPPG

import pandas as pd
from pyPPG.ppg_bm.bm_extraction import get_biomarkers

###########################################################################
##################### Get Biomarkers of PPG Derivatives ###################
###########################################################################
def get_ppg_derivs(s: pyPPG.PPG, fp: pyPPG.Fiducials):
    """
    This function returns the biomarkers of PPG derivatives.

    :param s: object of PPG signal
    :type s: pyPPG.PPG object
    :param fp: object of fiducial points
    :type fp: pyPPG.Fiducials object

    :return:
        - biomarkers: dictionary of biomarkers of PPG derivatives
        - biomarkers_lst: list a biomarkers with name, definition and unit
    """

    biomarkers_lst = [
                    ["Tu",       "u-point time, the time between the pulse onset and u-point", "[s]"],
                    ["Tv",       "v-point time, the time between the pulse onset and v-point", "[s]"],
                    ["Tw",       "w-point time, the time between the pulse onset and w-point", "[s]"],
                    ["Ta",       "a-point time, the time between the pulse onset and a-point", "[s]"],
                    ["Tb",       "b-point time, the time between the pulse onset and b-point", "[s]"],
                    ["Tc",       "c-point time, the time between the pulse onset and c-point", "[s]"],
                    ["Td",       "d-point time, the time between the pulse onset and d-point", "[s]"],
                    ["Te",       "e-point time, the time between the pulse onset and e-point", "[s]"],
                    ["Tf",       "f-point time, the time between the pulse onset and f-point", "[s]"],
                    ["Tb-c",	 "b-c time, the time between the b-point and c-point", "[s]"],
                    ["Tb-d",	 "b-d time, the time between the b-point and d-point", "[s]"],
                    ["Tp1",	     "p1-point time, the time between the pulse onset and p1-point", "[s]"],
                    ["Tp2",      "p2-point time, the time between the pulse onset and p2-point", "[s]"],
                    ["Tp1-dp",   "p1-dia time, the time between the p1-point and diastolic peak", "[s]"],
                    ["Tp2-dp",   "p2-dia time, the time between the p2-point and diastolic peak", "[s]"],
    ]

    header = ['name', 'definition', 'unit']
    biomarkers_lst = pd.DataFrame(biomarkers_lst, columns=header)

    df, df_biomarkers = get_biomarkers(s, fp, biomarkers_lst.name)
    
    return df_biomarkers, biomarkers_lst



import numpy as np
from dotmap import DotMap
from scipy.signal import filtfilt
from scipy import signal

def Preprocessing(s: DotMap, filtering: bool):
    '''The function calculates the preprocessed PPG, PPG', PPG", and PPG'".

    :param s: a struct of PPG signal:
        - s.v: a vector of PPG values
        - s.fs: the sampling frequency of the PPG in Hz
    :type s: DotMap

    :return: filt_sig, filt_d1, filt_d2, filt_d3: preprocessed PPG, PPG', PPG", and PPG'"
    '''

    ## PPG filtering
    if filtering:
        fL = 0.5
        fH = 12
        order = 4
        b,a = signal.cheby2(order, 20, [fL, fH], 'bandpass', fs=s.fs)
        filt_sig_cb2 = filtfilt(b, a, s.v)

        if s.fs >= 75:
            win = round(s.fs * 0.02)
            B = 1 / win * np.ones(win)
            filt_sig = filtfilt(B, 1, filt_sig_cb2)
        else:
            filt_sig=filt_sig_cb2
    else:
        filt_sig=s.v

    if s.fs >= 150:
        ## PPG' filtering
        win = round(s.fs * 0.01)
        B1 = 1 / win * np.ones(win)
        dx = np.gradient(s.v)
        filt_d1 = filtfilt(B1, 1, dx)

        ## PPG" filtering
        win = round(s.fs * 0.02)
        B2 = 1 / win * np.ones(win)
        dx = np.gradient(s.v)
        dx = filtfilt(B, 1, dx)
        ddx = np.gradient(dx)
        filt_d2 = filtfilt(B2, 1, ddx)

        ## PPG'" filtering
        filt_d3 = np.gradient(ddx)
    else:
        filt_d1 = np.gradient(filt_sig)
        filt_d2 = np.gradient(filt_d1)
        filt_d3 = np.gradient(filt_d2)

    return filt_sig, filt_d1, filt_d2, filt_d3


from pyPPG.pack_ppg._ErrorHandler import _check_shape_, WrongParameter
import pandas as pd
class PPG:
    '''
    This is class for the input PPG parameters.
    '''
    def __init__(self,s):
        """
        :param s: a struct of PPG signal:

            * s.start: beginning of the signal in sample
            * s.end: end of the signal in sample
            * s.v: a vector of PPG values
            * s.fs: the sampling frequency of the PPG in Hz
            * s.name: name of the record
            * s.v: 1-d array, a vector of PPG values
            * s.fs: the sampling frequency of the PPG in Hz
            * s.filt_sig: 1-d array, a vector of the filtered PPG values
            * s.filt_d1: 1-d array, a vector of the filtered PPG' values
            * s.filt_d2: 1-d array, a vector of the filtered PPG" values
            * s.filt_d3: 1-d array, a vector of the filtered PPG'" values
        :type s: DotMap

        """

        if s.fs <= 0:
            raise WrongParameter("Sampling frequency should be strictly positive")
        _check_shape_(s.v, s.fs)

        keys=s.keys()
        keys_list = list(keys)
        for i in keys_list:
            exec('self.'+i+' = s[i]')

    def get_s(self):
        """
        This function retrieves the struct of the PPG signal.

        :return: s: DataFrame of the PPG signal
        """
        keys = self.__dict__.keys()
        keys_list = list(keys)
        s = {}
        for i in keys_list:
            s[i] = getattr(self, i, None)

        return pd.DataFrame(s)


class Fiducials:
        '''
        This is class for the PPG fiducial points.
        '''
        def __init__(self, fp):
            """
            :param fiducials: a DataFrame where the key is the name of the fiducial pints and the value is the list of fiducial points PPG Fiducials Points.

                * PPG signal (fp.on, fp.sp, fp.dn, fp.dp): List of pulse onset, systolic peak, dicrotic notch, diastolic peak
                * 1st derivative (fp.u, fp.v, fp.w): List of points of 1st maximum and minimum in 1st derivitive between the onset to onset intervals
                * 2nd derivative (fp.a, fp.b, fp.c, fp.d, fp.e, fp.f): List of maximum and minimum points in 2nd derivitive between the onset to onset intervals
                * 3rd derivative (fp.p1, fp.p2): List of points of 1st maximum and minimum in 3rd derivitive between the onset to onset intervals
            :type fiducials: DataFrame

            """

            keys = fp.keys()
            keys_list = list(keys)
            for i in keys_list:
                exec('self.' + i + ' = fp[i]')

        def get_fp(self):
            """
            This function retrieves the struct of the fiducial points.

            :return: fp: DataFrame of the fiducial points
            """

            keys = self.__dict__.keys()
            keys_list = list(keys)
            fp={}
            for i in keys_list:
                fp[i]=getattr(self, i, None)

            return pd.DataFrame(fp)

        def get_row(self, row_index):
            """
            This function retrieves the specified row from the DataFrame of fiducial points.
            :param row_index: the index corresponding to the row in the fiducial points DataFrame

            :return: the corresponding row in the fiducial points DataFrame
            """
            keys = self.__dict__.keys()
            keys_list = list(keys)
            fp_row={}
            for i in keys_list:
                fp_row[i]=getattr(self, i, None)[row_index]

            return pd.DataFrame(fp_row,[row_index])

class Biomarkers:
        '''
        This is class for the PPG biomarkers.
        '''
        def __init__(self, bm_defs, bm_vals, bm_stats):
            """
                :param bm_defs: dictionary with name, definition and unit of biomarkers in different categories:

                    * def_ppg_sig: description of the PPG signal
                    * def_sig_ratios: description of the Signal ratios
                    * def_ppg_derivs: description of the PPG derivatives
                    * def_derivs_ratios: description of the Derivatives ratios
                :type bm_defs: dict
                :param bm_vals: dictionary with values of biomarkers in different categories:

                    * bm_ppg_sig: biomarkers of the PPG signal
                    * bm_sig_ratios: biomarkers of the Signal ratios
                    * bm_ppg_derivs: biomarkers of the PPG derivatives
                    * bm_derivs_ratios: biomarkers of the Derivatives ratios
                :type bm_vals: dict
                :param bm_stats: data frame with summary of PPG features

                    * def_ppg_sig: description of the PPG signal
                    * def_sig_ratios: description of the Signal ratios
                    * def_ppg_derivs: description of the PPG derivatives
                    * def_derivs_ratios: description of the Derivatives ratios
                :type bm_stats: dict

            """

            for i in ['bm_defs', 'bm_vals', 'bm_stats']:
                exec('self.' + i + ' = ' +i)

        def get_bm(self):
            """
            This function retrieves the dictionary of the biomarkers.

            :return: bm: dictionary of the biomarkers
            """

            keys = self.__dict__.keys()
            keys_list = list(keys)
            bm={}
            for i in keys_list:
                bm[i]=getattr(self, i, None)

            return pd.DataFrame(bm)



from setuptools import setup, find_packages
import io

with io.open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='pyPPG',
    version='1.0.13',
    description='pyPPG: a python toolbox for PPG morphological analysis.',
    author='Marton A. Goda, PhD',
    author_email="marton.goda@campus.technion.ac.il",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/godamartonaron/GODA_pyPPG",
    project_urls={"Bug Tracker": "https://github.com/pypa/sampleproject/issues",},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: OS Independent",
    ],
    packages={"pyPPG", "pyPPG/ppg_bm", "pyPPG/pack_ppg"},
    package_data={
        "pyPPG": ["*"],
        "pyPPG/ppg_bm": ["*"],
        "pyPPG/pack_ppg": ["*"],
    },

    install_requires=["numpy", "mne"],
    python_requires=">=3.10",
    include_package_data=True,
)


import numpy as np
import pandas as pd
import scipy.stats

###########################################################################
###################### Statistics of PPG Biomarkers #######################
###########################################################################

def get_statistics(peaks: pd.Series, onsets: pd.Series, ppg_biomarkers: dict):
    """
    The function compares the different biomedical features of PPG signal.

    :param peaks: 1-d array, peaks of the signal
    :type peaks: Series
    :param onsets: 1-d array, onsets of the signal
    :type onsets: Series
    :param ppg_biomarkers: dictionary of the PPG biomarkers
    :type ppg_biomarkers: dict

    :return: df_windows: data frame with summary of PPG features
    """

    BM_keys = list(ppg_biomarkers.keys())
    ppg_statistics={}
    for j in range (0, len(BM_keys)):
        df_features=ppg_biomarkers[BM_keys[j]]

        df_stat = pd.DataFrame()
        for k in df_features.keys():
            data=df_features[k].values
            df_tempstat = {}
            df_tempstat['mean'] = np.mean(data)
            df_tempstat['median'] = np.median(data)
            df_tempstat['std'] = np.std(data)
            percentile_25 = np.percentile(data, 25)
            percentile_75 = np.percentile(data, 75)
            df_tempstat['percentile_25'] = percentile_25
            df_tempstat['percentile_75'] = percentile_75
            df_tempstat['iqr'] = percentile_75 - percentile_25
            df_tempstat['skew'] = scipy.stats.skew(data)
            df_tempstat['kurtosis'] = scipy.stats.kurtosis(data)
            df_tempstat['mad'] = np.mean(np.absolute(data - np.mean(data)))
            df_stat[k]=df_tempstat.values()

        df_stat.index = ['mean', 'median', 'std', 'percentile_25', 'percentile_75', 'iqr', 'skew', 'kurtosis', 'mad']
        ppg_statistics[BM_keys[j]]=df_stat

    return ppg_statistics


import numpy as np
from scipy.signal import detrend, find_peaks, correlate

def get_ppgSQI(ppg: list, fs: int, annotation: list):
    '''
    PPG Signal Quality Index based on beat template correlation.

    :param ppg: a vector of PPG values
    :type ppg: int
    :param fs: Samples frequency
    :type fs: int
    :param annotation: PPG annotation time(samples)
    :type annotation: list

    :return: psqi: PPG Signal Quality Index


    Reference:
    ----------

    Original Matlab implementation:
    Qiao Li, November 10, 2014.
    https://github.com/MIT-LCP/PhysioNetChallengePublic/blob/master/2015/sample-submission/ppgSQI.m

    Author:
    Marton A. Goda – Faculty of Biomedical Engineering,
    Technion – Israel Institute of Technology, Haifa, Israel (August 2022)
    '''

    Fs = fs
    # Create PPG template
    t,v = use_template(ppg, annotation-1, fs)

    c1 = np.empty(len(annotation)-1)
    c1[:] = np.NaN
    psqi = np.empty(len(annotation)-1)
    psqi[:] = np.NaN

    for j in range (0,len(annotation) - 1):
        # Calculate correlation coefficients based on the template length
        beatbegin = annotation[j]-1
        beatend = annotation[j + 1]-1
        if beatend - beatbegin > 3 * Fs:
            beatend = beatbegin + 3 * Fs

        templatelength = len(t)
        if (beatbegin + templatelength - 1 > len(ppg)) or (beatend > len(ppg)) or (beatbegin < 1):
            continue

        currentb = j
        cc = np.corrcoef(t, ppg[beatbegin:beatbegin + templatelength])
        c1[j] = cc[0, 1]
        if (c1[j] < 0):
            c1[j] = 0

        psqi[currentb] = c1[j]

    return psqi

def use_template(wave, annotation: list, fs: int):
    '''
    PPG waveform template creation.
    Written by Qiao Li, February 21, 2011

    :param wave: a vector of PPG wave
    :type ppg: int
    :param fs: Samples frequency
    :type fs: int
    :param annotation: PPG annotation time(sample)
    :type ann_ppg: list

    :return:
        - template: PPG waveform template based on normal - length beats
        - valid: 1 for valid template, 0 for invalid template
    '''

    template = []
    valid = 0

    # according to heart rate max(300 bpm) and min(20 bpm) to get max and min beat - by - beat interval
    hr_max = 300
    bb_interval_min = fs * 60 / hr_max
    hr_min = 20
    bb_interval_max = fs * 60 / hr_min

    # Normal beat thresholds
    normal_beat_length_min = 0.7
    normal_beat_lentth_max = 1.5
    normal_beat_percent_threshold = 0.5

    # using correlate to get the basic period of the PPG as the length of template
    data = detrend(wave)

    y = correlate(data, data, 'full', method='fft')
    lenw = len(wave)
    lena = len(annotation)
    i = lenw

    locs = find_peaks(y[i:])[0]
    pks = y[locs]
    if len(pks)==0:
        return

    max_v=max(pks)
    max_i=np.where(pks==max_v)[0][0]
    i = locs[max_i]

    cycle = fs
    if i < lenw - 1:
        cycle = i

    # cumulate the beats with reasonable length to get template
    if lena < 2:
        return

    p0 = 1
    i = annotation[p0]

    temp_ahead=0
    while i - temp_ahead < 1:
        p0 = p0 + 1
        if (p0 > lena):
            template = wave
            valid = 0
            return

        i = annotation[p0]

    if p0 + 1 >= lena:
        return

    beat_interval = np.diff(annotation[p0:len(annotation)])
    median_bi = np.median(beat_interval)

    if median_bi!='NaN':
        temp_peak = abs(locs - median_bi)
        m = min(temp_peak)
        i = np.where(temp_peak==m)[0][0]
        cycle = locs[i]+1
    else:
        return

    # the length of template valid detection
    valid = 1
    if (cycle > bb_interval_max) or (cycle < bb_interval_min):
        valid = 0
        template = np.zeros(cycle)
        return

    n = 0
    d1 = 0
    invalidn = 0
    currentbeatlength = annotation[p0 + 1] - annotation[p0]
    if currentbeatlength > 0:
        d1 = wave[i - temp_ahead:i + cycle]
        n = 1
    else:
        invalidn = invalidn + 1
        d1 = np.zeros(cycle + temp_ahead)

    p0 = p0
    if p0 < lena - 1:
        i = annotation[p0]
        n = 1
        invalidn = 0
        while (i < lenw - cycle) and (p0 < lena - 1):
            currentbeatlength = annotation[p0 + 1] - annotation[p0]
            if currentbeatlength > 0:
                d1 = d1 + wave[i - temp_ahead:i + cycle]
                n = n + 1
            else:
                invalidn = invalidn + 1

            p0 = p0 + 1
            i = annotation[p0]

        d1 = d1/n
        # normal beat is less than the reasonable percentage of all beats
        if (n / (n + invalidn)) < normal_beat_percent_threshold:
            valid = 0
        else:
            valid = 0
    else:
        valid=0

    template = d1

    return template,valid